from traceutils2.scamper.hop cimport Hop, Trace, Reader, ICMPType

cdef class WartsHop(Hop):
    cdef:
        public list icmpext
        public int probe_id
        public unsigned short probe_size
        public int reply_ipid
        public int icmp_q_ipl
        public int icmp_nhmtu


cdef class WartsTrace(Trace):
    cdef:
        public str type
        public str version
        public userid
        public str method
        public int icmp_sum
        public str stop_reason
        public int stop_data
        public dict start
        public int hop_count
        public int attempts
        public unsigned char hoplimit
        public unsigned char firsthop
        public double wait
        public int wait_probe
        public int tos
        public unsigned char probe_size
        public unsigned int probe_count
        public int dport
        public int sport
        public str rtr

cdef class WartsPing:
    cdef:
        public str type, version, method, src, dst
        public dict start, statistics
        public int ping_sent, probe_size, userid, ttl, wait_us
        public double wait, timeout
        public list responses
        public int family
        public int dport
        public int sport
        public int tcp_seq
        public int tcp_ack
        public list flags
        public list probe_tsps

cdef list create_responses(list responses, int family);

cdef class WartsPingResponse:
    cdef:
        public int seq, reply_size, reply_ttl, probe_ipid, reply_ipid, icmp_type, icmp_code
        public str src, reply_proto
        public dict tx, rx
        public double rtt
        public int family
        public ICMPType type
        public list tsandaddr
        public list RR

cdef class AbstractWartsReader(Reader):
    cdef:
        bint trace
        bint ping
        public str hostname
        public str addr
        f
        public dict firstline
        bint safe
        bint parallel_read

    cdef void set_hostname(self) except *;

cdef class WartsReader(AbstractWartsReader):
    pass

cdef class WartsJsonReader(AbstractWartsReader):
    pass

# cdef class WartsReader(Reader):
#     cdef:
#         bint trace
#         bint ping
#         str hostname
#
#     cpdef void open(self) except *;
#     cpdef void close(self) except *;
#
# cdef class WartsJsonReader(Reader):
#     cdef:
#         bint trace
#         bint ping
#         str hostname
#         f
#
#     cpdef void open(self) except *;
#     cpdef void close(self) except *;



from file2 import fopen
from pb_amarder import Progress
from traceutils2.as2org.as2org cimport AS2Org
from traceutils2.radix.ip2as cimport IP2AS

from bdrmapit.graph.node cimport Interface, Router


cdef class Graph:
    def __init__(self, dict interfaces=None, dict routers=None):
        if interfaces is None:
            self.interfaces = {}
        else:
            self.interfaces = interfaces
        if routers is None:
            self.routers = {}
        else:
            self.routers = routers


# @cython.nonecheck(False)
# @cython.overflowcheck(False)
cpdef Graph construct_graph(list addrs, dict nexthop, dict multi, dict dps, list mpls, IP2AS ip2as, AS2Org as2org, str nodes_file=None, int increment=100000):
    cdef dict interfaces = {}, routers = {}
    cdef str addr, edge, nid
    cdef int asn, i, predcount
    cdef Interface interface, succ
    cdef Router router
    cdef list edges, dests, naddrs
    cdef set origins

    interfaces = {}
    routers = {}
    if nodes_file is not None:
        pb = Progress(message='Creating nodes', increment=increment)
        with fopen(nodes_file, 'rt') as f:
            for line in pb.iterator(f):
                if line[0] != '#':
                    _, nid, *naddrs = line.split()
                    nid = nid[:-1]
                    router = Router(nid)
                    routers[router.name] = router
                    for addr in naddrs:
                        asn = ip2as.asn(addr)
                        if asn > 0 or asn <= -100:
                            interface = Interface(addr, asn, as2org[asn])
                            interfaces[addr] = interface
                            interface.router = router
                            router.interfaces.append(interface)
                            interface.router = router
                            router.interfaces.append(interface)
                            routers[router.name] = router
    pb = Progress(len(addrs), 'Creating remaining routers and interfaces', increment=increment)
    for addr in pb.iterator(addrs):
        if nodes_file is None or addr not in interfaces:
            asn = ip2as.asn(addr)
            interface = Interface(addr, asn, as2org[asn])
            interfaces[addr] = interface
            router = Router(interface.addr)
            interface.router = router
            router.interfaces.append(interface)
            routers[router.name] = router
    pb = Progress(len(mpls), 'Noting MPLS interfaces', increment=increment)
    for addr in pb.iterator(mpls):
        interface = interfaces[addr]
        interface.mpls = True
    # for interface in interfaces.values():
    #     if not interface.router:
    #         router = Router(interface.addr)
    #         interface.router = router
    #         router.interfaces.append(interface)
    #         routers[router.name] = router
    pb = Progress(len(nexthop), 'Adding nexthop edges', increment=increment)
    for addr, edges in pb.iterator(nexthop.items()):
        interface = interfaces[addr]
        router = interface.router
        router.nexthop = True
        for i in range(len(edges)):
            edge = edges[i]
            succ = interfaces[edge]
            if succ in router.succ:
                origins = router.origins[succ]
                origins.add(interface.asn)
            else:
                router.succ.add(succ)
                router.origins[succ] = {interface.asn}
            predcount = succ.pred.get(router, 0)
            succ.pred[router] = predcount + 1
    pb = Progress(len(multi), 'Adding multihop edges', increment=increment)
    for addr in pb.iterator(multi):
        interface = interfaces[addr]
        router = interface.router
        if not router.nexthop:
            edges = multi[addr]
            for edge in edges:
                succ = interfaces[edge]
                if succ in router.succ:
                    router.origins[succ].add(interface.asn)
                else:
                    router.succ.add(succ)
                    router.origins[succ] = {interface.asn}
    pb = Progress(len(dps), 'Adding destination ASes', increment=increment)
    for addr, dests in pb.iterator(dps.items()):
        interface = interfaces[addr]
        interface.dests.update(dests)
    return Graph(interfaces=interfaces, routers=routers)

    # for i in range(len(addrs)):
    #     addr = addrs[i]
    #     asn = ip2as.asn(addr)
    #     interface = Interface(addr, asn, as2org[asn])
    #     interfaces[addr] = interface
    # for addr in mpls:
    #     interface = interfaces[addr]
    #     interface.mpls = True
    # for interface in interfaces.values():
    #     if not interface.router:
    #         router = Router(interface.addr)
    #         interface.router = router
    #         router.interfaces.append(interface)
    #         routers[router.name] = router
    # for addr, edges in nexthop.items():
    #     interface = interfaces[addr]
    #     router = interface.router
    #     router.nexthop = True
    #     for i in range(len(edges)):
    #         edge = edges[i]
    #         succ = interfaces[edge]
    #         if succ in router.succ:
    #             origins = router.origins[succ]
    #             origins.add(interface.asn)
    #         else:
    #             router.succ.add(succ)
    #             router.origins[succ] = {interface.asn}
    #         predcount = succ.pred.get(router, 0)
    #         succ.pred[router] = predcount + 1
    # for addr in multi:
    #     interface = interfaces[addr]
    #     router = interface.router
    #     if not router.nexthop:
    #         edges = multi[addr]
    #         for edge in edges:
    #             succ = interfaces[edge]
    #             if succ in router.succ:
    #                 router.origins[succ].add(interface.asn)
    #             else:
    #                 router.succ.add(succ)
    #                 router.origins[succ] = {interface.asn}
    # for addr, dests in dps.items():
    #     interface = interfaces[addr]
    #     interface.dests.update(dests)
    # return Graph(interfaces=interfaces, routers=routers)



cimport cython
from socket import inet_ntop
from libc.string cimport strchr, strtok
from libc.stdlib cimport atoi


cdef class RadixPrefix:
    # def __init__(self, bytes addr=None, unsigned char bitlen=0, unsigned char family=0):
    #     self.addr = addr
    #     self.bitlen = bitlen
    #     self.family = family

    def __repr__(self):
        return self.network()

    cpdef str network(self):
        cdef addr = inet_ntop(self.family, self.addr)
        return '{}/{}'.format(addr, self.bitlen)


@cython.cdivision(True)
@cython.initializedcheck(False)
cdef bytes fix4(bytes a, unsigned char masklen):
    cdef unsigned char c[20]
    cdef unsigned char i, quotient, remainder, mask
    quotient = masklen / 8
    remainder = masklen % 8
    inet_pton(AF_INET, a, c)
    for i in range(quotient+1, 4, 1):
        c[i] = 0
    # if remainder > 0:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask
    return <bytes>c[:4]


@cython.cdivision(True)
@cython.initializedcheck(False)
cdef bytes fix6(bytes a, unsigned char masklen):
    cdef unsigned char c[20]
    cdef unsigned char i, quotient, remainder, mask
    # print('reached')
    quotient = masklen / 8
    remainder = masklen % 8
    inet_pton(AF_INET6, a, c)
    for i in range(quotient+1, 16, 1):
        c[i] = 0
    # if remainder > 0:
    mask = ((~0) << (8 - remainder))
    c[quotient] &= mask
    # print(test)
    # test = <bytes>c[:16]
    return bytes(<bytes>c[:16])
    # return bytes(test)


cdef bytes inet_fix(unsigned char family, bytes a, unsigned char masklen):
    if family == AF_INET:
        return fix4(a, masklen)
    return fix6(a, masklen)


cdef unsigned char find_family(bytes network):
    if strchr(network, b':'):
        return AF_INET6
    return AF_INET


cdef (char*, char*) partition(bytes b):
    cdef char *net
    cdef char *mask
    net = strtok(b, b'/')
    if net != NULL:
        mask = strtok(NULL, b'/')
        if mask == NULL:
            mask = b''
    else:
        net = b''
        mask = b''
    return net, mask


def fake():
    pass


cdef RadixPrefix from_network(str s, char masklen=-1):
    cdef bytes a = s.encode()
    cdef bytes net, mask
    net, mask = partition(a)
    # print(net, mask)
    # net, _, mask = a.partition(b'/')
    # cdef Network net = partition(a)
    family = find_family(net)
    # print(family)
    if mask:
        if masklen >= 0:
            raise ValueError('masklen is set twice')
        masklen = atoi(mask)
    elif masklen < 0:
        if family == AF_INET:
            masklen = 32
        else:
            masklen = 128
    cdef bytes addr = inet_fix(family, net, masklen)
    # if family == AF_INET6:
    #     print(addr, len(addr))
    cdef RadixPrefix prefix = RadixPrefix()
    prefix.addr = addr
    prefix.bitlen = masklen
    prefix.family = family
    return prefix


cdef RadixPrefix from_address(str s):
    cdef bytes a = s.encode()
    cdef unsigned char c[16]
    cdef unsigned char family = find_family(a)
    cdef RadixPrefix prefix = RadixPrefix()
    inet_pton(family, a, c)
    if family == AF_INET:
        prefix.addr = <bytes>c[:4]
        prefix.bitlen = 32
    else:
        prefix.addr = <bytes>c[:16]
        prefix.bitlen = 128
    prefix.family = family
    return prefix


cdef RadixPrefix from_packed(bytes addr, char masklen=-1):
    cdef RadixPrefix prefix = RadixPrefix()
    prefix.addr = addr
    if len(addr) == 4:
        if masklen >= 0:
            prefix.bitlen = masklen
        else:
            prefix.bitlen = 32
        prefix.family = AF_INET
    else:
        if masklen >= 0:
            prefix.bitlen = masklen
        else:
            prefix.bitlen = 128
        prefix.family = AF_INET6
    return prefix



from traceutils2.utils.dicts cimport StrDict, EmptyDict

cdef class AS2Org:
    cdef:
        public StrDict orgs
        public StrDict asn_names
        public StrDict asn_org_names
        readonly EmptyDict siblings
        public dict org_cc

    cpdef str name(self, int asn);
    cpdef str asn_name(self, int asn);
    cpdef str asn_cc(self, int asn);


NOTIMPLEMENTED = 0
NODEST = 1
MODIFIED = 3
SINGLE = 4
SINGLE_MODIFIED = 5
HEAPED = 6
HEAPED_MODIFIED = 7
MISSING_NOINTER = 10
MISSING_INTER = 9

REALLOCATED_PREFIX = 500
REALLOCATED_DEST = 1000
SINGLE_SUCC_ORIGIN = 10
SINGLE_SUCC_4 = 11
SUCC_ORIGIN_INTER = 12
SUCC_ORIGIN_CUST = 13
REMAINING_4 = 14
IUPDATE = 15
ALLPEER_SUCC = 16
ALLPEER_ORIGIN = 17
IASN_SUCC_HALF = 18
ALLRELS = 19
VOTE_SINGLE = 50
VOTE_TIE = 70
SINGLE_SUCC_RASN = 15
HIDDEN_INTER = 100
HIDDEN_NOINTER = 200



import sys
from collections import Counter
from typing import Collection, List, Dict

from pb_amarder import Progress
from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.utils.utils import max_num, peek

from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Router, Interface


DEBUG = False

NOTIMPLEMENTED = 0
NODEST = 1
MODIFIED = 3
SINGLE = 4
SINGLE_MODIFIED = 5
HEAPED = 6
HEAPED_MODIFIED = 7
MISSING_NOINTER = 10
MISSING_INTER = 9

REALLOCATED_PREFIX = 500
REALLOCATED_DEST = 1000
SINGLE_SUCC_ORIGIN = 10
SINGLE_SUCC_4 = 11
SUCC_ORIGIN_INTER = 12
SUCC_ORIGIN_CUST = 13
REMAINING_4 = 14
IUPDATE = 15
ALLPEER_SUCC = 16
ALLPEER_ORIGIN = 17
IASN_SUCC_HALF = 18
ALLRELS = 19
VOTE_SINGLE = 50
VOTE_TIE = 70
SINGLE_SUCC_RASN = 15
HIDDEN_INTER = 100
HIDDEN_NOINTER = 200


class Debug:

    def __init__(self):
        self.old = DEBUG

    def __enter__(self):
        global DEBUG
        DEBUG = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        global DEBUG
        DEBUG = self.old
        return False


class Bdrmapit:

    def __init__(self, graph: Graph, as2org: AS2Org, bgp: BGP):
        self.graph = graph
        self.as2org = as2org
        self.bgp = bgp
        self.rupdates = Updates()
        self.iupdates = Updates()
        self.routers_succ: List[Router] = []
        self.lasthops: List[Router] = []
        for router in graph.routers.values():
            if router.succ:
                self.routers_succ.append(router)
            else:
                self.lasthops.append(router)
        self.interfaces_pred: List[Interface] = [i for i in graph.interfaces.values() if i.pred]
        self.previous_updates = []

    def annotate_router(self, router: Router):
        isucc: Interface
        if DEBUG:
            print('Nexthop: {}'.format(router.nexthop))
        asns = Counter(i.asn for i in router.interfaces if i.asn > 0)
        if DEBUG: print('IASNS: {}'.format(asns))
        for isucc in router.succ:
            if DEBUG: print('Succ: {}'.format(isucc))
            iupdate = self.iupdates[isucc]
            rupdate = self.rupdates[isucc.router]
            if DEBUG:
                print('\tIUPDATE: {}'.format(iupdate))
                print('\tRUPDATE: {}'.format(rupdate))
            if iupdate and iupdate.asn > 0:
                if not rupdate:
                    asn = isucc.asn
                else:
                    if rupdate.asn > 0 and isucc.org == rupdate.org:
                        asn = iupdate.asn
                    else:
                        asn = isucc.asn
            else:
                asn = isucc.asn
            if asn > 0:
                asns[asn] += 1
        if not asns:
            return 0
        selections = max_num(asns, key=lambda x: asns[x])
        if len(selections) == 1:
            return selections[0]
        return router.interfaces[0].asn

    def annotate_routers(self, routers: Collection[Router], increment=100000):
        pb = Progress(len(routers), 'Annotating routers', increment=increment)
        for router in pb.iterator(routers):
            asn = self.annotate_router(router)
            self.rupdates.add_update(router, asn, self.as2org[asn], 1)

    def annotate_interface(self, interface: Interface):
        edges: Dict[Router, int] = interface.pred
        asns = Counter()
        if interface.asn > 0:
            asns[interface.asn] += 1
        if DEBUG: print('IASN: {}'.format(asns))
        for rpred in edges:
            rasn = self.rupdates.asn(rpred)
            if DEBUG:
                print('Pred={addr:} RASN={rasn:} {num:}'.format(addr=rpred.name, rasn=rasn, num=len(rpred.succ)))
            if rasn > 0:
                asns[rasn] += 1
        selections = max_num(asns, key=lambda x: asns[x])
        if len(selections) == 1:
            return selections[0], len(edges)
        return interface.asn, 0

    def annotate_interfaces(self, interfaces: Collection[Interface]):
        pb = Progress(len(interfaces), 'Adding links', increment=100000)
        for interface in pb.iterator(interfaces):
            if interface.asn >= 0:
                asn, utype = self.annotate_interface(interface)
                self.iupdates.add_update(interface, asn, self.as2org[asn], utype)

    def annotate_stubs(self, routers: Collection[Router]):
        pb = Progress(len(routers), 'Stub Heuristic', increment=100000)
        for router in pb.iterator(routers):
            if len(router.succ) == 1:
                interface = router.interfaces[0]
                iupdate = self.iupdates[interface]
                if iupdate and iupdate.utype > 1:
                    if interface.org != iupdate.org:
                        continue
                isucc = peek(router.succ)
                iupdate = self.iupdates[isucc]
                if iupdate and iupdate.utype > 1:
                    # if isucc.org != iupdate.org:
                        continue
                rupdate = self.rupdates[isucc.router]
                if rupdate:
                    if isucc.org != rupdate.org:
                        continue
                conesize = self.bgp.conesize[isucc.asn]
                if conesize < 5 and conesize < self.bgp.conesize[router.interfaces[0].asn]:
                    self.rupdates.add_update_direct(router, isucc.asn, isucc.org, 2)

    def annotate_lasthops(self, routers: List[Router]):
        pb = Progress(len(routers), increment=1000000)
        for router in pb.iterator(routers):
            interface = router.interfaces[0]
            self.rupdates.add_update_direct(router, interface.asn, interface.org, 0)

    def graph_refinement(self, routers: List[Router], interfaces: List[Interface], iterations=-1):
        iteration = 0
        while iterations < 0 or iteration < iterations:
            Progress.message('********** Iteration {:,d} **********'.format(iteration), file=sys.stderr)
            self.annotate_routers(routers)
            self.rupdates.advance()
            self.annotate_interfaces(interfaces)
            self.iupdates.advance()
            ru = dict(self.rupdates)
            iu = dict(self.iupdates)
            if (ru, iu) in self.previous_updates:
                break
            self.previous_updates.append((ru, iu))
            iteration += 1



from collections import defaultdict, Counter
from typing import Optional

from pb_amarder import Progress
from traceutils2.as2org.as2org import AS2Org
from traceutils2.bgp.bgp import BGP
from traceutils2.utils.utils import peek, max_num

from bdrmapit.algorithm import debug
from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Interface


class FirstHopMixin:
    as2org: Optional[AS2Org] = None
    bgp: Optional[BGP] = None
    graph: Optional[Graph] = None
    caches: Optional[Updates] = None
    iupdates: Optional[Updates] = None

    def annotate_firsthop(self, interface: Interface, votes):
        if len(votes) == 1:
            asn = peek(votes)
        else:
            asns = max_num(votes, key=votes.__getitem__)
            if debug.DEBUG: print('MaxNum: {}'.format(asns))
            rels = [asn for asn in asns if interface.asn == asn or self.bgp.rel(interface.asn, asn)]
            if not rels:
                rels = asns
            if debug.DEBUG:
                print('Rels: {}'.format(rels))
                print('Sorted Rels: {}'.format(sorted(rels, key=lambda x: (
                    x != interface.asn, -self.bgp.provider_rel(interface.asn, x), -self.bgp.conesize[x], x))))
            asn = min(rels, key=lambda x: (x != interface.asn, -self.bgp.conesize[x], x))
        return asn

    def annotate_firsthops(self, firstcounters, filemap):
        prevs = defaultdict(Counter)
        pb = Progress(len(firstcounters), 'Creating previous ASNs', increment=100000, callback=lambda: '{:,d}'.format(len(prevs)))
        for file, addr in pb.iterator(firstcounters):
            if file in filemap:
                prevs[addr][filemap[file]] += 1
        pb = Progress(len(prevs), 'Annotating first hops', increment=100000, callback=lambda: '{:,d}'.format(len(prevs)))
        for addr, files in pb.iterator(prevs.items()):
            interface = self.graph.interfaces[addr]
            if addr in prevs and self.iupdates.asn(interface) == -1:
                asn = self.annotate_firsthop(interface, prevs[addr])
                self.caches.add_update_direct(interface, asn, self.as2org[asn], 3)



from bdrmapit.graph.node cimport Node

cdef class UpdateObj:
    cdef public int asn, utype
    cdef public str org

cdef class Updates(dict):
    cdef public str name
    cdef public dict changes

    cpdef void add_update(self, Node node, int asn, str org, int utype) except *;
    cpdef void add_update_direct(self, Node node, int asn, str org, int utype) except *;
    cpdef void advance(self) except *;
    cpdef int asn(self, node) except *;
    cpdef Updates make_copy(self, str name=*);
    cpdef str org(self, Node node);

cdef class UpdatesView(Updates):
    pass
    


import pickle
from collections import defaultdict

from bdrmapit.algorithm.parse_results_container import Container


class ContainerLoop(Container):
    def __init__(self, ip2as, as2org, adjs=None, loop=None, **kwargs):
        super().__init__(ip2as, as2org, **kwargs)
        self.adjs = adjs
        self.loop = loop

    def loadloop(self, file):
        with open(file, 'rb') as f:
            loopinfo = pickle.load(f)
        self.adjs = loopinfo['adjs']
        self.loop = loopinfo['loop']

    def remove_loops(self, prunedict, edgedict):
        for x, ys in prunedict.items():
            if x in edgedict:
                newys = [y for y in edgedict[x] if y not in ys]
                if not newys:
                    del edgedict[x]
                else:
                    edgedict[x] = newys

    def prune_loopinfo(self):
        prune = {t for t, n in self.loop.items() if n / self.adjs[t] >= .5}
        prunedict = defaultdict(set)
        for x, y in prune:
            prunedict[x].add(y)
        self.remove_loops(prunedict, self.nexthop)
        self.remove_loops(prunedict, self.multi)

    def prune_addrs(self):
        oldaddrs = set(self.addrs)
        self.addrs = {a for t, n in self.adjs.items() if t not in self.loop or self.loop[t] / n < .5 for a in t if a in oldaddrs}
        for x, ys in self.nexthop.items():
            self.addrs.add(x)
            self.addrs.update(ys)
        for x, ys in self.multi.items():
            self.addrs.add(x)
            self.addrs.update(ys)



cdef class Router:

    def __init__(self, str name):
        self.name = name
        self.interfaces = []
        self.nexthop = False
        self.vrf = False
        self.succ = set()
        self.dests = set()
        self.origins = {}
        # self.origins = defaultdict(set)
        self.hints = None

    def __repr__(self):
        return 'Router<{}>'.format(self.name)

    cpdef Router copy(self):
        router = Router(self.name)
        router.interfaces.extend(self.interfaces)
        router.nexthop = self.nexthop
        router.vrf = self.vrf
        router.succ.update(self.succ)
        router.dests.update(self.dests)
        router.origins.update(self.origins)
        router.hints = self.hints
        return router

cdef class Interface:

    def __init__(self, str addr, int asn, str org):
        self.addr = addr
        self.asn = asn
        self.org = org
        self.router = None
        self.pred = {}
        self.dests = set()
        self.vrf = False
        # self.mpls = False
        self.hint = 0

    def __repr__(self):
        return 'Interface<{} {}>'.format(self.addr, self.asn)

    cpdef Interface copy(self):
        iface = Interface(self.addr, self.asn, self.org)
        iface.router = self.router
        iface.pred.update(self.pred)
        iface.dests.update(self.dests)
        iface.vrf = self.vrf
        iface.hint = self.hint
        return iface



IF UNAME_SYSNAME == 'Windows':
    cdef extern from '<ws2tcpip.h>':
        cdef int AF_INET, AF_INET6
        cdef int inet_pton(int, char*, void*)
ELSE:
    cdef extern from '<arpa/inet.h>':
        cdef int AF_INET, AF_INET6
        cdef int inet_pton(int, char*, void*)

cdef class RadixPrefix:
    cdef readonly bytes addr
    cdef readonly unsigned char bitlen
    cdef readonly unsigned char family

    cpdef str network(self);

cdef bytes fix4(bytes a, unsigned char masklen);
cdef bytes fix6(bytes a, unsigned char masklen);
cdef bytes inet_fix(unsigned char family, bytes a, unsigned char masklen);
cdef unsigned char find_family(bytes network);
# cdef Network partition(bytes b);
cdef RadixPrefix from_network(str s, char masklen=*);
cdef RadixPrefix from_address(str s);
cdef RadixPrefix from_packed(bytes addr, char masklen=*);
cdef (char*, char*) partition(bytes b);



from traceutils2.radix.radix cimport Radix
from traceutils2.radix.radix_node cimport RadixNodeASN


cdef list PRIVATE4, PRIVATE6
cdef str MULTICAST4, MULTICAST6


cdef class IP2AS(Radix):
    cpdef RadixNodeASN add_asn(self, str network, short masklen=*, long asn=*);
    cpdef long asn(self, str asn);
    cpdef long asn_packed(self, bytes packed);
    cpdef void add_private(self) except *

cpdef IP2AS create_private();
cpdef IP2AS create_table(str filename);



import bz2
import gzip
import orjson as json
from subprocess import Popen, PIPE

from file2 import fopen2, fopen
from traceutils2.scamper.hop cimport Hop, Trace, Reader, gettype
from traceutils2.utils.net cimport find_family

cdef list create_hops(list hops, int family):
        cdef dict h
        if hops is not None:
            return [WartsHop(family=family, **h) for h in hops]
        return []


cdef class WartsTrace(Trace):

    def __init__(
            self, str type='', str version='', userid=-1, str method='', str src='', str dst='',
            int icmp_sum=-1, str stop_reason='', int stop_data=-1, dict start=None, int hop_count=-1,
            int attempts=-1, unsigned char hoplimit=0, unsigned char firsthop=1, double wait=-1,
            int wait_probe=-1, int tos=-1, unsigned short probe_size=0, unsigned int probe_count=0,
            list hops=None, str list_name='', int id=-1, str hostname='', long start_time=0,
            int dport=0, int sport=0, str rtr=None, str jdata=None
    ):
        self.src = src
        self.dst = dst
        self.family = find_family(dst.encode())
        self.allhops = create_hops(hops, self.family)
        self.hops = self.allhops
        self.jdata = jdata

        self.type = type
        self.version = version
        self.userid = userid
        self.method = method
        self.icmp_sum = icmp_sum
        self.stop_reason = stop_reason
        self.stop_data = stop_data
        self.start = start
        self.hop_count = hop_count
        self.attempts = attempts
        self.hoplimit = hoplimit
        self.firsthop = firsthop
        self.wait = wait
        self.wait_probe = wait_probe
        self.tos = tos
        self.probe_size = probe_size
        self.probe_count = probe_count
        self.dport = dport
        self.sport = sport
        self.rtr = rtr


cdef class WartsHop(Hop):

    def __init__(
            self, str addr='', unsigned char probe_ttl=0, int probe_id=-1, unsigned short probe_size=0,
            double rtt=-1, unsigned char reply_ttl=0, int reply_tos=-1, unsigned short reply_size=0,
            int reply_ipid=-1, unsigned char icmp_type=0, unsigned char icmp_code=0, unsigned char icmp_q_ttl=1,
            int icmp_q_ipl=-1, int icmp_q_tos=-1, list icmpext=None, int icmp_nhmtu=-1, int family=0, **kwargs
    ):
        self.addr = addr
        self.probe_ttl = probe_ttl
        self.rtt = rtt
        self.reply_ttl = reply_ttl
        self.reply_tos = reply_tos
        self.reply_size = reply_size
        self.icmp_type = icmp_type
        self.icmp_code = icmp_code
        self.icmp_q_ttl = icmp_q_ttl
        self.icmp_q_tos = icmp_q_tos
        self.icmpext = icmpext
        if icmpext:
            self.ismpls = 'mpls_labels' in icmpext[0]
        else:
            self.ismpls = False
        # self.num = iptoint(addr)

        self.probe_id = probe_id
        self.probe_size = probe_size
        self.reply_ipid = reply_ipid
        self.icmp_q_ipl = icmp_q_ipl
        self.icmp_nhmtu = icmp_nhmtu
        self.family = family
        self.type = gettype(family, icmp_type, icmp_code)


cdef class WartsPing:
    def __init__(
            self, str type='ping', str version=None, str method=None, str src=None, str dst=None, dict start=None,
            int ping_sent=-1, int probe_size=-1, int userid=-1, int ttl=-1, double wait=-1, int wait_us=-1,
            double timeout=-1, list responses=None, dict statistics=None,
            int dport = 0, int sport = 0, int tcp_seq = 0, int tcp_ack = 0, list flags = None, list probe_tsps = None
    ):
        self.type = type
        self.version = version
        self.method = method
        self.src = src
        self.dst = dst
        self.family = find_family(dst.encode())
        self.start = start
        self.ping_sent = ping_sent
        self.probe_size = probe_size
        self.userid = userid
        self.ttl = ttl
        self.wait = wait
        self.timeout = timeout
        self.responses = create_responses(responses, self.family)
        self.statistics = statistics
        self.dport = dport
        self.sport = sport
        self.tcp_seq = tcp_seq
        self.tcp_ack = tcp_ack
        self.flags = flags
        self.probe_tsps = probe_tsps

    def __repr__(self):
        result = ['Src={src}, Dst={dst}:'.format(src=self.src, dst=self.dst)]
        for resp in self.responses:
            result.append('\t{}'.format(repr(resp)))
        result.append(repr(self.statistics))
        return '\n'.join(result)


cdef list create_responses(list responses, int family):
    cdef dict resp
    resps = []
    for resp in responses:
        resp['src'] = resp.pop('from', None)
        resps.append(WartsPingResponse(family=family, **resp))
    return resps


cdef class WartsPingResponse:
    def __init__(
            self, str src=None, int seq=-1, int reply_size=-1, int reply_ttl=-1, str reply_proto=None, dict tx=None, dict rx=None,
            double rtt=-1, int probe_ipid=-1, int reply_ipid=-1, int icmp_type=-1, int icmp_code=-1, int family=0,
            list tsandaddr = None, RR=None
    ):
        self.src = src
        self.seq = seq
        self.reply_size = reply_size
        self.reply_ttl = reply_ttl
        self.reply_proto = reply_proto
        self.tx = tx
        self.rx = rx
        self.rtt = rtt
        self.probe_ipid = probe_ipid
        self.reply_ipid = reply_ipid
        self.icmp_type = icmp_type
        self.icmp_code = icmp_code
        self.family = family
        self.tsandaddr = tsandaddr
        self.RR = RR
        self.type = gettype(family, icmp_type, icmp_code)

    def __repr__(self):
        return 'RTT={rtt}'.format(rtt=self.rtt)

cdef class AbstractWartsReader(Reader):
    def __init__(self, str filename, bint trace=True, bint ping=True, bint safe=True, bint parallel_read=False):
        self.filename = filename
        self.f = None
        self.trace = trace
        self.ping = ping
        self.hostname = None
        self.addr = None
        self.firstline = None
        self.safe = safe
        self.parallel_read = parallel_read

    def safe_iter(self):
        cdef:
            str line
        if not self.safe:
            return self.f
        fiter = iter(self.f)
        while True:
            try:
                line = next(fiter)
                yield line
            except StopIteration:
                break
            except EOFError as e:
                print(e, self.filename)

    def __iter__(self):
        cdef:
            str line, rtype
            dict j
        if self.firstline is not None:
            rtype = self.firstline['type']
            if rtype == 'trace' and self.trace:
                yield WartsTrace(jdata=str(self.firstline), **self.firstline)
            elif rtype == 'ping' and self.ping:
                yield WartsPing(**self.firstline)
        for line in self.safe_iter():
            j = json.loads(line)
            rtype = j['type']
            if rtype == 'trace':
                if self.trace:
                    yield WartsTrace(jdata=line, **j)
            elif rtype == 'ping':
                if self.ping:
                    yield WartsPing(**j)

    def json(self):
        cdef:
            str line
        if self.firstline is not None:
            yield self.firstline
        for line in self.safe_iter():
            j = json.loads(line)
            yield j

    cdef void set_hostname(self) except *:
        cdef:
            str line
            dict j
        for line in self.f:
            # print(line)
            j = json.loads(line)
            if j['type'] == 'cycle-start':
                self.hostname = j['hostname']
                self.addr = j.get('addr', None)
            else:
                self.firstline = j
                # print(self.firstline)
            break

    def raw(self):
        if self.firstline is not None:
            yield self.firstline
        yield from self.safe_iter()

cdef class WartsReader(AbstractWartsReader):
    def __init__(self, str filename, bint trace=True, bint ping=True, bint safe=True, bint parallel_read=False):
        super().__init__(filename, trace=trace, ping=ping, parallel_read=parallel_read)
        self.p = None

    cpdef void open(self) except *:
        cdef str cmd
        if self.filename.endswith('.bz2') or self.filename.endswith('.bzip2'):
            cmd = 'bzip2 -d -c {} | sc_warts2json'
        elif self.filename.endswith('.gz'):
            cmd = 'gzip -d -c {} | sc_warts2json'
        else:
            cmd = 'sc_warts2json {}'
        self.p = Popen(cmd.format(self.filename), stdout=PIPE, shell=True, universal_newlines=True)
        self.f = self.p.stdout
        self.set_hostname()

    cpdef void close(self) except *:
        self.f.close()
        self.p.wait()
        self.p = None

cdef class WartsJsonReader(AbstractWartsReader):
    cpdef void open(self) except *:
        cdef str cmd
        if self.parallel_read:
            self.f = fopen2(self.filename, 'rt', processes=0)
        else:
            self.f = fopen2(self.filename, 'rt', processes=1)
        self.set_hostname()

    cpdef void close(self) except *:
        self.f.close()

# cdef class WartsJsonReader(AbstractWartsReader):
#     cpdef void open(self) except *:
#         cdef str cmd
#         if self.filename.endswith('.bz2') or self.filename.endswith('.bzip2'):
#             self.f = bz2.open(self.filename, 'rt')
#         elif self.filename.endswith('.gz'):
#             self.f = gzip.open(self.filename, 'rt')
#         else:
#             self.f = open(self.filename, 'rt')
#         self.set_hostname()
#
#     cpdef void close(self) except *:
#         self.f.close()

# cdef class WartsReader(Reader):
#     def __init__(self, str filename, bint trace=True, bint ping=True):
#         self.filename = filename
#         self.p = None
#         self.trace = trace
#         self.ping = ping
#         self.hostname = None
#         self.f = None
#
#     def __iter__(self):
#         cdef str line, rtype
#         cdef dict j
#         for line in self.p.stdout:
#             j = json.loads(line)
#             rtype = j['type']
#             if rtype == 'trace':
#                 if self.trace:
#                     yield WartsTrace(jdata=line, **j)
#             elif rtype == 'ping':
#                 if self.ping:
#                     yield WartsPing(**j)
#
#     def json(self):
#         for line in self.p.stdout:
#             j = json.loads(line)
#             yield j
#
#     def raw(self):
#         return self.p.stdout
#
#     cpdef void open(self) except *:
#         cdef str cmd
#         if self.filename.endswith('.bz2') or self.filename.endswith('.bzip2'):
#             cmd = 'bzip2 -d -c {} | sc_warts2json'
#         elif self.filename.endswith('.gz'):
#             cmd = 'gzip -d -c {} | sc_warts2json'
#         else:
#             cmd = 'sc_warts2json {}'
#         self.p = Popen(cmd.format(self.filename), stdout=PIPE, shell=True, universal_newlines=True)
#         for line in self.f:
#             break
#         j = json.loads(line)
#         if j['type'] == 'cycle-start':
#             self.hostname = j['hostname']
#
#     cpdef void close(self) except *:
#         self.p.stdout.close()
#         self.p.wait()
#         self.p = None
#
# cdef class WartsJsonReader(Reader):
#     def __init__(self, str filename, bint trace=True, bint ping=True):
#         self.filename = filename
#         self.f = None
#         self.trace = trace
#         self.ping = ping
#         self.hostname = None
#
#     def __iter__(self):
#         cdef str line, rtype
#         cdef dict j
#         for line in self.f:
#             j = json.loads(line)
#             rtype = j['type']
#             if rtype == 'trace':
#                 if self.trace:
#                     try:
#                         yield WartsTrace(jdata=line, **j)
#                     except:
#                         print(line)
#                         raise
#             elif rtype == 'ping':
#                 if self.ping:
#                     yield WartsPing(**j)
#
#     def json(self):
#         for line in self.f:
#             j = json.loads(line)
#             yield j
#
#     def raw(self):
#         return self.f
#
#     cpdef void open(self) except *:
#         cdef str cmd
#         if self.filename.endswith('.bz2') or self.filename.endswith('.bzip2'):
#             self.f = bz2.open(self.filename, 'rt')
#         elif self.filename.endswith('.gz'):
#             self.f = gzip.open(self.filename, 'rt')
#         else:
#             self.f = open(self.filename, 'rt')
#         for line in self.f:
#             break
#         j = json.loads(line)
#         if j['type'] == 'cycle-start':
#             self.hostname = j['hostname']
#         else:
#             self.f.seek(0)
#
#     cpdef void close(self) except *:
#         self.f.close()



import json
import os.path
import sqlite3
from sqlite3 import Connection

from pb_amarder import Progress

from bdrmapit.algorithm.algorithm import Bdrmapit
from bdrmapit.graph.node import Router, Interface

def srows(router: Router, add_preds=False, add_origins=False):
    rows = []
    for s in router.succ:
        row = {'addr': s.addr, 'router': s.router.name}
        if add_preds:
            row['pred'] = prows(s, add_succs=False)
        if add_origins:
            row['origins'] = list(router.origins[s])
        rows.append(row)
    return rows

def prows(iface: Interface, add_succs=False):
    rows = []
    for p, num in iface.pred.items():
        row = {'router': p.name, 'num': num}
        if add_succs:
            row['succ'] = srows(p, add_preds=False)
        rows.append(row)
    return rows

def insert(con: Connection, values):
    con.executemany('INSERT INTO graph (addr, graph) VALUES (:addr, :graph)', values)
    con.commit()
    values.clear()

def savegraph(bdrmapit: Bdrmapit, filename, replace=True):
    exists = os.path.exists(filename)
    if not exists or replace:
        if exists:
            os.remove(filename)
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(dir_path, 'graph.sql')) as f:
            script = f.read()
        con = sqlite3.connect(filename)
        cur = con.cursor()
        cur.executescript(script)
        con.commit()
    else:
        con = sqlite3.connect(filename)
    values = []
    pb = Progress(len(bdrmapit.graph.interfaces), increment=100000)
    for interface in pb.iterator(bdrmapit.graph.interfaces.values()):
        router: Router = interface.router
        succs = json.dumps(srows(router, add_preds=True, add_origins=True))
        preds = json.dumps(prows(interface, add_succs=True))
        row = {'addr': interface.addr, 'router': router.name, 'router_info': succs, 'iface_info': preds}
        values.append(row)
        if len(values) >= 100000:
            insert(con, values)
    if values:
        insert(con, values)



#!/usr/bin/env python
import pickle
from argparse import ArgumentParser
from collections import Counter, defaultdict
from enum import Enum
from multiprocessing.pool import Pool
from typing import Optional, List, Dict

from file2 import fopen
from pb_amarder import Progress
from traceutils2.radix.ip2as import IP2AS, create_table
from traceutils2.scamper.atlas import AtlasReader
from traceutils2.scamper.hop import ICMPType, Hop
from traceutils2.scamper.warts import WartsReader, WartsJsonReader

_ip2as: Optional[IP2AS] = None
_filemap4: Optional[Dict[str, str]] = None
_filemap6: Optional[Dict[str, str]] = None

class OutputType(Enum):
    WARTS = 1
    ATLAS = 2
    ATLAS_ODD = 3
    JSONWARTS = 4

class TraceFile:
    def __init__(self, filename, type):
        self.filename = filename
        self.type = type

    def __repr__(self):
        return self.filename

class ParseResults:

    def __init__(self):
        self.addrs = set()
        self.dps = set()
        self.spoofing = set()
        self.echos = set()
        self.cycles = set()
        self.loopadjs = Counter()
        self.nextadjs = Counter()
        self.multiadjs = Counter()
        self.first = Counter()
        # self.triplets = Counter()

    def __repr__(self):
        return 'Addrs {addrs:,d} N {nhop:,d} M {multi:,d} DPs {dests:,d} S {spoof:,d} E {echo:,d} C {cycle:,d} L {loop:,d} F {first:,d}'.format(
            addrs=len(self.addrs), nhop=len(self.nextadjs), multi=len(self.multiadjs), dests=len(self.dps),
            spoof=len(self.spoofing), echo=len(self.echos), cycle=len(self.cycles), loop=len(self.loopadjs),
            first=len(self.first),
            # triplets=len(self.triplets)
        )

    def __str__(self):
        return self.__repr__()

    def dump(self, file):
        with open(file, 'wb') as f:
            pickle.dump(vars(self), f)

    @classmethod
    def load(cls, file):
        with open(file, 'rb') as f:
            d = pickle.load(f)
        results = cls()
        for k in d:
            if hasattr(results, k):
                getattr(results, k).update(d[k])
        return results

    def update(self, results):
        for k, v in vars(results).items():
            getattr(self, k).update(v)

def parse(tfile: TraceFile):
    # public_ip4 = _filemap4.get(tfile.filename)
    # public_ip6 = _filemap6.get(tfile.filename)
    results: ParseResults = ParseResults()
    if tfile.type == OutputType.WARTS:
        f = WartsReader(tfile.filename, ping=False)
    elif tfile.type == OutputType.ATLAS:
        f = AtlasReader(tfile.filename)
    elif tfile.type == OutputType.JSONWARTS:
        f = WartsJsonReader(tfile.filename)
    else:
        raise Exception('Invalid output type: {}.'.format(tfile.type))
    f.open()
    if f.addr:
        public_ip4 = f.addr
        public_ip6 = f.addr
    else:
        public_ip4 = _filemap4[f.hostname] if f.hostname in _filemap4 else _filemap4.get(tfile.filename)
        public_ip6 = _filemap6[f.hostname] if f.hostname in _filemap6 else _filemap6.get(tfile.filename)
    try:
        fiter = iter(f)
        while True:
            try:
                trace = next(fiter)
                trace.prune_private(_ip2as)
                trace.prune_dups()
                trace.prune_loops(True)
                if trace.loop:
                    results.cycles.update(trace.loop)
                hops: List[Hop] = [h for h in trace.hops if _ip2as[h.addr] != -1 and h.addr != trace.src and h.addr != public_ip4 and h.addr != public_ip6]
                if not hops: continue
                fhop: Hop = hops[0]
                if fhop.probe_ttl == 1:
                    results.first[tfile.filename, fhop.addr] += 1
                lhop: Hop = hops[-1]
                if lhop.type == ICMPType.echo_reply or lhop.type == ICMPType.portping:
                    results.echos.add(lhop.addr)
                dst_asn = _ip2as.asn(trace.dst)
                for i in range(len(hops)):
                    x: Hop = hops[i]
                    results.addrs.add(x.addr)
                    if x.type != ICMPType.echo_reply and x.type != ICMPType.portping:
                        results.dps.add((x.addr, dst_asn))
                    if i == len(hops) - 1:
                        break
                    y: Hop = hops[i+1]
                    if y.type == ICMPType.echo_reply or y.type == ICMPType.portping:
                        break
                    if y.type == ICMPType.spoofing and y.icmp_q_ttl > 1:
                        break
                    distance = y.probe_ttl - x.probe_ttl
                    if y.icmp_q_ttl == 0:
                        distance += 1
                    if distance > 1:
                        distance = 2
                    elif distance < 1:
                        distance = -1
                    if y.type == ICMPType.spoofing:
                        results.spoofing.add((x.addr, y.addr, distance))
                    else:
                        if distance == 1:
                            results.nextadjs[x.addr, y.addr] += 1
                        else:
                            results.multiadjs[x.addr, y.addr] += 1
                if trace.loop:
                    for x, y in zip(trace.loop, trace.loop[1:]):
                        results.loopadjs[x.addr, y.addr] += 1
            except UnicodeDecodeError:
                print(tfile.filename, 'UnicodeDecodeError')
                break
            except EOFError:
                print(tfile.filename, 'EOFError')
                break
            except StopIteration:
                break
    finally:
        f.close()
    return results

def parse_sequential(files):
    results = ParseResults()

    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: str(results))
    for tfile in pb.iterator(files):
        newresults = parse(tfile)
        results.update(newresults)
    return results

def parse_parallel(files, poolsize):
    results = ParseResults()

    pb = Progress(len(files), 'Parsing traceroute files', callback=lambda: str(results))
    with Pool(poolsize) as pool:
        for newresults in pb.iterator(pool.imap_unordered(parse, files)):
            results.update(newresults)
    return results

def run(files, ip2as: IP2AS, poolsize, output=None, filemap4=None, filemap6=None):
    global _ip2as, _filemap4, _filemap6
    _ip2as = ip2as
    _filemap4 = filemap4 if filemap4 is not None else {}
    _filemap6 = filemap6 if filemap6 is not None else {}

    poolsize = min(len(files), poolsize)
    print(poolsize)
    results = parse_parallel(files, poolsize) if poolsize != 1 else parse_sequential(files)
    if output:
        results.dump(output)
    return results

def read_filemap(filename):
    filemap = {}
    with open(filename) as f:
        for line in f:
            if line[0] == '#':
                continue
            file, addr = line.split()
            filemap[file] = addr
    return filemap

def set_parser(parser: ArgumentParser, group=None, output=True):
    oparser = parser
    if group is not None:
        parser = parser.add_argument_group(group)
    parser.add_argument('-w', '--wfiles', help='File with list of newline-separated warts filenames.')
    parser.add_argument('-W', '--wfilelist', nargs='+', help='List of warts filenames, space separated.')
    parser.add_argument('-a', '--afiles', help='File with list of newline-separated atlas filenames.')
    parser.add_argument('-A', '--afilelist', nargs='+', help='List of filenames, space separated.')
    parser.add_argument('-j', '--jfiles')
    parser.add_argument('-J', '--jfilelist', nargs='+')
    oparser.add_argument('-i', '--ip2as', required=True, help='Filename of prefix-to-AS mappings in CAIDA prefix2as format.')
    parser.add_argument('-p', '--poolsize', type=int, default=1)
    parser.add_argument('-m', '--filemap4', help='Mapping from filename to public IPv4 address (tab separated).')
    parser.add_argument('-M', '--filemap6', help='Mapping from filename to public IPv4 address (tab separated).')
    if output:
        parser.add_argument('-o', '--output', required=True, help='Filename for pickle output file.')

def main(args=None, ip2as=None):
    if args is None:
        parser = ArgumentParser()
        set_parser(parser)
        args = parser.parse_args()
    files = []
    if args.wfiles:
        with fopen(args.wfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.WARTS) for line in f if line[0] != '#')
    if args.wfilelist:
        files.extend(TraceFile(file, OutputType.WARTS) for file in args.wfilelist)
    if args.afiles:
        with fopen(args.afiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.ATLAS) for line in f if line[0] != '#')
    if args.afilelist:
        files.extend(TraceFile(file, OutputType.ATLAS) for file in args.afilelist)
    if args.jfiles:
        with fopen(args.jfiles) as f:
            files.extend(TraceFile(line.strip(), OutputType.JSONWARTS) for line in f if line[0] != '#')
    if args.jfilelist:
        files.extend(TraceFile(file, OutputType.JSONWARTS) for file in args.jfilelist)
    filemap4 = read_filemap(args.filemap4) if args.filemap4 else {}
    filemap6 = read_filemap(args.filemap6) if args.filemap6 else {}
    if ip2as is None:
        ip2as = create_table(args.ip2as)
    return run(files, ip2as, args.poolsize, args.output, filemap4=filemap4, filemap6=filemap6)

if __name__ == '__main__':
    main()



cimport cython

from traceutils2.radix.radix_node cimport RadixNode, RadixNodeASNs
from traceutils2.radix.radix_prefix cimport RadixPrefix


# @cython.boundscheck(False)
# @cython.overflowcheck(False)
cdef unsigned char addr_test(bytes addr, unsigned char bitlen):
    cdef unsigned char left, right, bitand, bitshift
    bitshift = bitlen >> 3
    bitand = bitlen & 0x07
    left = addr[bitshift]
    right = 0x80 >> bitand
    return left & right


@cython.cdivision(True)
# @cython.boundscheck(False)
# @cython.overflowcheck(False)
cdef bint prefix_match(RadixPrefix left, RadixPrefix right, unsigned char bitlen) except -1:
    cdef bytes laddr, raddr
    cdef unsigned char quotient, remainder, mask, bitmainder, lmask, rmask, biteq
    laddr = left.addr
    raddr = right.addr
    if laddr is None or raddr is None:
        return False
    try:
        quotient = bitlen / 8
        for i in range(quotient):
            if laddr[i] != raddr[i]:
                return False
    except TypeError:
        # print(laddr, raddr, type(laddr) == type(bytes), type(raddr) == type(bytes))
        return False
        # raise
    remainder = bitlen % 8
    if remainder == 0:
        return True
    bitmainder = 8 - remainder
    mask = (~0) << bitmainder
    # mask = (~0) << (8 - remainder)
    lmask = laddr[quotient] & mask
    rmask = raddr[quotient] & mask
    biteq = lmask == rmask
    return biteq


cdef class RadixTree:

    def __init__(self):
        # self.maxbits = 128
        self.head = None
        self.active_nodes = 0

    cdef RadixNode create_node(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None):
        return RadixNode(prefix, prefix_size, parent)

    cdef RadixNode add(self, RadixPrefix prefix):
        cdef RadixNode head, parent, new_node, glue_node
        cdef bytes addr, test_addr
        cdef unsigned char bitlen, check_bit, differ_bit, i, r, j

        node = self.head
        if node is None:
            # easy case
            node = self.create_node(prefix)
            self.head = node
            self.active_nodes += 1
            return node
        addr = prefix.addr
        bitlen = prefix.bitlen
        # find the best place for the node
        while node.bitlen < bitlen or node.prefix is None:
            if addr_test(addr, node.bitlen):
                if node.right is None:
                    break
                node = node.right
            else:
                if node.left is None:
                    break
                node = node.left
        # find the first differing bit
        test_addr = node.prefix.addr
        check_bit = node.bitlen if node.bitlen < bitlen else bitlen
        differ_bit = 0
        i = 0
        while i * 8 < check_bit:
            r = addr[i] ^ test_addr[i]
            if r == 0:
                differ_bit = (i + 1) * 8
                i += 1
                continue
            # bitwise check
            for j in range(8):
                if r & (0x80 >> j):
                    break
            differ_bit = i * 8 + j
            break
        if differ_bit > check_bit:
            differ_bit = check_bit
        # now figure where to insert
        parent = node.parent
        while parent and parent.bitlen >= differ_bit:
            node, parent = parent, node.parent
        # found a match
        if differ_bit == bitlen and node.bitlen == bitlen:
            if not node.prefix:
                node.prefix = prefix
            return node
        # no match, new node
        new_node = self.create_node(prefix)
        self.active_nodes += 1
        # fix it up
        if node.bitlen == differ_bit:
            new_node.parent = node
            if addr_test(addr, node.bitlen):
                node.right = new_node
            else:
                node.left = new_node
            return new_node
        if bitlen == differ_bit:
            if addr_test(test_addr, bitlen):
                new_node.right = node
            else:
                new_node.left = node
            new_node.parent = node.parent
            if node.parent is None:
                self.head = new_node
            elif node.parent.right == node:
                node.parent.right = new_node
            else:
                node.parent.left = new_node
            node.parent = new_node
        else:
            glue_node = self.create_node(prefix=None, prefix_size=differ_bit, parent=node.parent)
            self.active_nodes += 1
            if addr_test(addr, differ_bit):
                glue_node.right = new_node
                glue_node.left = node
            else:
                glue_node.right = node
                glue_node.left = new_node
            new_node.parent = glue_node
            if node.parent is None:
                self.head = glue_node
            elif node.parent.right == node:
                node.parent.right = glue_node
            else:
                node.parent.left = glue_node
            node.parent = glue_node
        return new_node

    cdef void remove(self, RadixNode node) except *:
        cdef RadixNode parent, child

        if node.right and node.left:
            node.prefix.addr = None
            node.asn = 0
            node.bitlen = 0
            return
        if node.right is None and node.left is None:
            parent = node.parent
            self.active_nodes -= 1
            if parent is None:
                self.head = None
                return
            if parent.right == node:
                parent.right = None
                child = parent.left
            else:
                parent.left = None
                child = parent.right
            if parent.prefix:
                return
            # remove the parent too
            if parent.parent is None:
                self.head = child
            elif parent.parent.right == parent:
                parent.parent.right = child
            else:
                parent.parent.left = child
            child.parent = parent.parent
            self.active_nodes -= 1
            return
        if node.right:
            child = node.right
        else:
            child = node.left
        parent = node.parent
        child.parent = parent
        self.active_nodes -= 1

        if parent is None:
            self.head = child
            return
        if parent.right == node:
            parent.right = child
        else:
            parent.left = child
        return

    cdef RadixNode search_best(self, RadixPrefix prefix):
        cdef RadixNode node
        cdef bytes addr
        cdef unsigned char bitlen
        cdef list stack

        if self.head is None:
            return None
        node = self.head
        addr = prefix.addr
        bitlen = prefix.bitlen

        stack = []
        while node.bitlen < bitlen:
            if node.prefix:
                stack.append(node)
            if addr_test(addr, node.bitlen):
                node = node.right
            else:
                node = node.left
            if node is None:
                break
        if node and node.prefix:
            stack.append(node)
        # for node in reversed(stack):
        for i in range(len(stack) - 1, -1, -1):
            node = stack[i]
            if node.bitlen <= bitlen and prefix_match(node.prefix, prefix, node.bitlen):
                return node
        return None

    cdef RadixNode search_exact(self, RadixPrefix prefix):
        cdef RadixNode node
        cdef bytes addr
        cdef unsigned char bitlen

        if self.head is None:
            return None
        node = self.head
        addr = prefix.addr
        bitlen = prefix.bitlen

        while node.bitlen < bitlen:
            if addr_test(addr, node.bitlen):
                node = node.right
            else:
                node = node.left
            if node is None:
                return None

        if node.bitlen > bitlen or node.prefix is None:
            return None

        if prefix_match(node.prefix, prefix, bitlen):
            return node
        return None

    cdef RadixNode search_worst(self, RadixPrefix prefix):
        cdef RadixNode node
        cdef bytes addr
        cdef unsigned char bitlen

        if self.head is None:
            return None
        node = self.head
        addr = prefix.addr
        bitlen = prefix.bitlen

        stack = []
        while node.bitlen < bitlen:
            if node.prefix:
                stack.append(node)
            if addr_test(addr, node.bitlen):
                node = node.right
            else:
                node = node.left
            if node is None:
                break
        if node and node.prefix:
            stack.append(node)
        if len(stack) <= 0:
            return None
        for node in stack:
            if prefix_match(node.prefix, prefix, node.bitlen):
                return node
        return None

    cdef list search_covered(self, RadixPrefix prefix):
        cdef list results, stack
        cdef RadixNode node
        cdef bytes addr
        cdef unsigned char bitlen

        results = []
        if self.head is None:
            return results
        node = self.head
        addr = prefix.addr
        bitlen = prefix.bitlen

        while node.bitlen < bitlen:
            if addr_test(addr, node.bitlen):
                node = node.right
            else:
                node = node.left
            if node is None:
                return results

        stack = [node]
        while stack:
            node = stack.pop()
            if prefix_match(node.prefix, prefix, prefix.bitlen):
                results.append(node)
            if node.right:
                stack.append(node.right)
            if node.left:
                stack.append(node.left)

        return results

cdef class RadixTreeASN(RadixTree):

    cdef RadixNode create_node(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None):
        return RadixNodeASN(prefix, prefix_size, parent)

    cdef RadixNodeASN add_asn(self, RadixPrefix prefix, int asn):
        cdef RadixNodeASN node = <RadixNodeASN> self.add(prefix)
        node.asn = asn
        return node

cdef class RadixTreeASNs(RadixTree):

    cdef RadixNode create_node(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None):
        return RadixNodeASNs(prefix, prefix_size, parent)

    cdef RadixNodeASNs add_asns(self, RadixPrefix prefix, list asns):
        cdef RadixNodeASNs node = <RadixNodeASNs> self.add(prefix)
        node.asns = asns
        return node

cdef class RadixTreeData(RadixTree):

    cdef RadixNode create_node(self, RadixPrefix prefix=None, unsigned char prefix_size=0, RadixNode parent=None):
        return RadixNodeData(prefix, prefix_size, parent)

    cdef RadixNodeData add_data(self, RadixPrefix prefix, dict data=None):
        cdef RadixNodeData node = <RadixNodeData> self.add(prefix)
        if data is not None:
            node.data.update(data)
        return node



from traceutils2.radix.radix_prefix cimport RadixPrefix

cdef class RadixNode:
    cdef public RadixPrefix prefix
    cdef RadixNode parent, left, right
    cdef public unsigned char bitlen

cdef class RadixNodeASN(RadixNode):
    cdef public int asn

cdef class RadixNodeASNs(RadixNode):
    cdef public list asns

cdef class RadixNodeData(RadixNode):
    cdef public dict data



from file2 import fopen

from traceutils2.radix.radix cimport Radix
from traceutils2.radix.radix_node cimport RadixNodeData
from traceutils2.radix.radix_tree cimport RadixTreeData
from libc.stdlib cimport atol

cdef list PRIVATE4 = ['0.0.0.0/8', '10.0.0.8/8', '100.64.0.0/10', '127.0.0.0/8', '169.254.0.0/16', '172.16.0.0/12',
            '192.0.0.0/24', '192.0.2.0/24', '192.31.196.0/24', '192.52.193.0/24', '192.88.99.0/24', '192.168.0.0/16',
            '192.175.48.0/24', '198.18.0.0/15', '198.51.100.0/24', '203.0.113.0/24', '240.0.0.0/4',
            '255.255.255.255/32']
cdef list PRIVATE6 = ['::1/128', '::/128', '::ffff:0:0/96', '64:ff9b::/96', '100::/64', '2001::/23', '2001::/32', '2001:1::1/128',
            '2001:2::/48', '2001:3::/32', '2001:4:112::/48', '2001:5::/32', '2001:10::/28', '2001:20::/28',
            '2001:db8::/32', '2002::/16', '2620:4f:8000::/48', 'fc00::/7', 'fe80::/10']
cdef str MULTICAST4 = '224.0.0.0/3'
cdef str MULTICAST6 = 'FF00::/8'

cdef class IP2Data(Radix):

    def __init__(self):
        tree4 = RadixTreeData()
        tree6 = RadixTreeData()
        super().__init__(tree4, tree6)

    def __getitem__(self, str item):
        return self.data(item)

    cpdef RadixNodeData add_data(self, str network, short masklen=-1, dict data=None):
        cdef RadixNodeData node = self.add(network, masklen)
        if data is not None:
            node.data.update(data)
        return node

    cpdef dict data(self, str addr):
        cdef RadixNodeData node = self.search_best(addr)
        if node:
            return node.data
        return {}

    cpdef dict data_packed(self, bytes packed):
        cdef RadixNodeData node = self.search_best_packed(packed, -1)
        if node:
            return node.data
        return {}

    cpdef void add_private(self) except *:
        cdef dict data = {'asn': -1}
        for prefix in PRIVATE4:
            self.add_data(prefix, -1, data)
        for prefix in PRIVATE6:
            self.add_data(prefix, -1, data)
        self.add_data(MULTICAST4, -1, data)
        self.add_data(MULTICAST6, -1, data)


cpdef IP2Data create_private():
    cdef IP2Data ip2as = IP2Data()
    ip2as.add_private()
    return ip2as


cpdef IP2Data create_table(str filename):
    cdef IP2Data ip2as
    cdef bytes prefix, asn_str
    cdef long asn
    cdef str prefix_s
    ip2as = IP2Data()
    with fopen(filename, 'rb') as f:
        # f.readline()
        for line in f:
            prefix, asn_str = line.rstrip().split()
            asn = atol(asn_str)
            prefix_s = prefix.decode()
            ip2as.add_data(prefix_s, -1, data={'asn': asn})
    return ip2as



IF UNAME_SYSNAME == 'Windows':
    cdef extern from '<ws2tcpip.h>':
        # ctypedef int socklen_t
        ctypedef int size_t
        cdef size_t INET_ADDRSTRLEN, INET6_ADDRSTRLEN
        cdef int AF_INET, AF_INET6
        cdef int inet_pton(int, char*, void*)
        cdef char *inet_ntop(int, void*, char*, size_t)
ELSE:
    cdef extern from '<arpa/inet.h>':
        ctypedef int socklen_t
        cdef socklen_t INET_ADDRSTRLEN, INET6_ADDRSTRLEN
        cdef int AF_INET, AF_INET6
        cdef int inet_pton(int, char*, void*)
        cdef char *inet_ntop(int, void*, char*, socklen_t)

cdef bytes inet_pton4(bytes a);
cdef bytes inet_pton6(bytes a);
cdef bytes inet_pton_bytes(unsigned char family, bytes a);
cdef bytes inet_pton_str(unsigned char family, str s);
cdef bytes inet_pton_auto(bytes a);
cdef bytes inet_pton_auto_str(str s);
cpdef unsigned char find_family(bytes addr);
cdef void fix4(bytes a, unsigned char masklen, unsigned char *c) except *;
cdef void fix6(bytes a, unsigned char masklen, unsigned char *c) except *;
cpdef bytes inet_fix(unsigned char family, bytes a, unsigned char masklen);
cdef list prefix_addrs4(bytes addr, int size);
cdef list prefix_addrs6(bytes addr, int size);
cpdef list prefix_addrs(str addr, int size);
cdef bytes otherside4(bytes addr, int num);
cdef bytes otherside6(bytes addr, int num);
cpdef str otherside(str addr, int num);
cpdef bint valid(long asn) except -1;
cdef void fix4_bytes(bytes a, unsigned char masklen, unsigned char *c) except *;
cdef void fix6_bytes(bytes a, unsigned char masklen, unsigned char *c) except *;
cpdef bytes inet_fix_bytes(unsigned char family, bytes a, unsigned char masklen);



from enum import IntEnum


class VType(IntEnum):
    toforward = 1
    forwarding = 2
    both = 3


class VRFEdge:
    def __init__(self, node, vtype: VType):
        self.node = node
        self.vtype = vtype

    def update(self, vtype: VType):
        if vtype != self.vtype:
            self.vtype = VType.both



from traceutils2.scamper.hop cimport Hop, Trace, Reader

cdef extern from '<arpa/inet.h>':
    cdef int AF_INET, AF_INET6

cdef class AtlasHop(Hop):
    cdef:
        public dict icmpext
        public str flags
        public int late
        public int dup
        public str edst
        list hdropts

cdef class AtlasTrace(Trace):
    cdef:
        public int af
        public str dst_name, proto, type
        public long endtime, timestamp
        public list result

cdef class AtlasReader(Reader):
    cdef object f



import json
import os
import re
import sqlite3
from collections import Counter
from typing import Set, Collection

import pandas as pd
from file2 import fopen2, fopen
from pb_amarder import Progress

from traceutils2.radix.ip2as import IP2AS

from bdrmapit.algorithm.algorithm import Bdrmapit
from bdrmapit.algorithm.updates_dict import Updates, UpdateObj
from bdrmapit.graph.node import Interface, Router
from bdrmapit.scripts.traceparser import ParseResults


class Save:

    def __init__(self, filename, bdrmapit: Bdrmapit, rupdates: Updates = None, iupdates: Updates = None, replace=True):
        self.filename = filename
        self.bdrmapit = bdrmapit
        self.rupdates = rupdates if rupdates is not None else bdrmapit.rupdates
        self.iupdates = iupdates if iupdates is not None else bdrmapit.iupdates
        exists = os.path.exists(filename)
        if not exists or replace:
            if exists:
                os.remove(filename)
            dir_path = os.path.dirname(os.path.realpath(__file__))
            with open(os.path.join(dir_path, 'tables.sql')) as f:
                script = f.read()
            con = sqlite3.connect(filename)
            cur = con.cursor()
            cur.executescript(script)
            con.commit()
            con.close()

    def save_annotations(self):
        interface: Interface
        values = []
        pb = Progress(len(self.bdrmapit.graph.interfaces), 'Writing annotations', increment=100000)
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        for interface in pb.iterator(self.bdrmapit.graph.interfaces.values()):
            addr = interface.addr
            router: Router = interface.router
            rupdate: UpdateObj = self.rupdates[router]
            iupdate: UpdateObj = self.iupdates[interface]
            if rupdate is None:
                rasn = -1
                rorg = -1
                rtype = -1
            else:
                rasn = rupdate.asn
                rorg = rupdate.org
                rtype = rupdate.utype
            if iupdate is None or interface.org != rorg:
                iasn = interface.asn
                iorg = interface.org
                itype = -1 if iupdate is None else 0
            else:
                iasn = iupdate.asn
                iorg = iupdate.org
                itype = iupdate.utype
            phop = bool(interface.pred)
            row = {'addr': addr, 'router': router.name, 'asn': rasn, 'org': rorg, 'conn_asn': iasn, 'conn_org': iorg, 'echo': False, 'nexthop': router.nexthop, 'phop': phop, 'rtype': rtype, 'itype': itype, 'iasn': interface.asn}
            # if addr == '202.68.67.250':
            #     print(row)
            values.append(row)
            if len(values) > 100000:
                cur.executemany('INSERT INTO annotation (addr, router, asn, org, conn_asn, conn_org, echo, nexthop, phop, rtype, itype, iasn) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :echo, :nexthop, :phop, :rtype, :itype, :iasn)', values)
                con.commit()
                values.clear()
        if values:
            cur.executemany('INSERT INTO annotation (addr, router, asn, org, conn_asn, conn_org, echo, nexthop, phop, rtype, itype, iasn) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :echo, :nexthop, :phop, :rtype, :itype, :iasn)', values)
            con.commit()
        cur.close()
        con.close()

    def save_echos(self, echos, ip2as, as2org):
        interface: Interface
        values = []
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        pb = Progress(len(echos), 'Writing echos', increment=100000)
        for addr in pb.iterator(echos):
            rasn = iasn = ip2as[addr]
            rorg = iorg = as2org[rasn]
            rtype = 0
            itype = 0
            row = {'addr': addr, 'router': addr, 'asn': rasn, 'org': rorg, 'conn_asn': iasn, 'conn_org': iorg, 'echo': True, 'rtype': rtype, 'itype': itype}
            values.append(row)
            if len(values) > 100000:
                cur.executemany(
                    'INSERT INTO annotation (addr, router, asn, org, conn_asn, conn_org, echo, rtype, itype) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :echo, :rtype, :itype)',
                    values)
                con.commit()
                values.clear()
        if values:
            cur.executemany('INSERT INTO annotation (addr, router, asn, org, conn_asn, conn_org, echo, rtype, itype) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :echo, :rtype, :itype)', values)
            con.commit()
        cur.close()
        con.close()

    def save_ixps(self):
        values = []
        for router in self.bdrmapit.routers_succ:
            conn_asn = self.rupdates[router].asn
            conn_org = self.bdrmapit.as2org[conn_asn]
            for isucc in router.succ:
                if isucc.asn <= -100:
                    pid = (isucc.asn * -1) - 100
                    rsucc = isucc.router
                    asn = self.rupdates[rsucc].asn
                    org = self.bdrmapit.as2org[asn]
                    value = {'addr': isucc.addr, 'router': router.name, 'asn': asn, 'org': org, 'conn_asn': conn_asn, 'conn_org': conn_org, 'pid': pid, 'nexthop': router.nexthop}
                    values.append(value)
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        cur.executemany('INSERT INTO ixp (addr, router, asn, org, conn_asn, conn_org, pid, nexthop) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :pid, :nexthop)', values)
        con.commit()
        con.close()

    def save_links(self):
        values = []
        for isucc in self.bdrmapit.interfaces_pred:
            rsucc = isucc.router
            asn = self.rupdates[rsucc].asn
            org = self.bdrmapit.as2org[asn]
            ixp = isucc.asn <= -100
            for router in isucc.pred:
                conn_asn = self.rupdates[router].asn
                conn_org = self.bdrmapit.as2org[conn_asn]
                if conn_org != org:
                    value = {'addr': isucc.addr, 'router': router.name, 'asn': asn, 'org': org, 'conn_asn': conn_asn, 'conn_org': conn_org, 'ixp': ixp}
                    values.append(value)
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        cur.executemany('INSERT INTO link (addr, router, asn, org, conn_asn, conn_org, ixp) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :ixp)', values)
        con.commit()
        con.close()

    def save_caches(self):
        values = []
        for isucc, iupdate in self.bdrmapit.caches.items():
            rsucc = isucc.router
            asn = self.rupdates[rsucc].asn
            org = self.bdrmapit.as2org[asn]
            ixp = isucc.asn <= -100
            conn_asn = iupdate.asn
            conn_org = iupdate.org
            if conn_org != org:
                value = {'addr': isucc.addr, 'router': rsucc.name, 'asn': asn, 'org': org, 'conn_asn': conn_asn, 'conn_org': conn_org, 'ixp': ixp}
                values.append(value)
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        cur.executemany('INSERT INTO cache (addr, router, asn, org, conn_asn, conn_org, ixp) VALUES (:addr, :router, :asn, :org, :conn_asn, :conn_org, :ixp)', values)
        con.commit()
        con.close()

    def extras(self, parseres: ParseResults, ip2as: IP2AS):
        values = []
        loops = set()
        for addrs in parseres.loopadjs:
            for addr in addrs:
                if addr not in self.bdrmapit.graph.interfaces:
                    loops.add(addr)
                    asn = ip2as[addr]
                    org = self.bdrmapit.as2org[asn]
                    row = {'addr': addr, 'asn': asn, 'reason': 'loop'}
                    values.append(row)
        echos = set()
        for addr in parseres.echos:
            if addr not in loops and addr not in self.bdrmapit.graph.interfaces:
                echos.add(addr)
                asn = ip2as[addr]
                org = self.bdrmapit.as2org[asn]
                row = {'addr': addr, 'asn': asn, 'reason': 'echo'}
                values.append(row)
        con = sqlite3.connect(self.filename)
        cur = con.cursor()
        cur.executemany('INSERT INTO excluded (addr, asn, org, reason) VALUES (:addr, :asn, :org, :reason)', values)
        con.commit()
        con.close()

    def save_node_as(self, filename, include_all=False):
        with fopen(filename, 'wt') as f:
            for router in self.bdrmapit.graph.routers.values():
                if router.name[0] == 'N':
                    update = self.bdrmapit.rupdates[router]
                    if update.asn <= 0:
                        continue
                    if update.utype == 1:
                        method = 'interfaces'
                    elif update.utype < 10:
                        method = 'last_hop'
                    else:
                        method = 'refinement'
                    f.write('node.AS {}:  {} {}\n'.format(router.name, update.asn, method))

def isregexasn(utype):
    return (utype & 0xff00) == 0xff00

def isregexorg(utype):
    return utype & 0xfe00 == 0xfe00

def regexreasons(utype):
    reasons = []
    if utype == 0xff00:
        reasons.append('noinfo')
    if utype & 0x0001:
        reasons.append('origin')
    if utype & 0x0002:
        reasons.append('subsequent')
    if utype & 0x0004:
        reasons.append('dest')
    if utype & 0x0008:
        reasons.append('provider')
    if not reasons:
        print(utype)
    return reasons

class ITDK:
    def __init__(self, bdrmapit: Bdrmapit, rupdates: Updates = None):
        self.bdrmapit = bdrmapit
        self.rupdates = rupdates if rupdates is not None else bdrmapit.rupdates

    def default_reason(self, router, update: UpdateObj):
        if update.asn <= 0:
            return 'unknown'
        elif router.hints and update.asn in router.hints:
            return 'as-hints'
        elif router.succ:
            return 'refinement'
        elif router.dests:
            return 'lasthop'
        else:
            return 'origins'

    def hint_reason(self, router, update: UpdateObj):
        utype = update.utype
        if isregexasn(utype):
            reasons = regexreasons(utype)
            reason = '|'.join(reasons)
        elif isregexorg(utype):
            reasons = regexreasons(utype)
            reason = '|'.join(reasons)
        else:
            reason = 'normal'
        return reason

    def write_nodes(self, filename, reason_func=None, include_all=False):
        if reason_func is None:
            reason_func = self.default_reason
        with fopen2(filename, 'wt') as f:
            for name, router in self.bdrmapit.graph.routers.items():
                if include_all or name[0] == 'N':
                    update = self.rupdates[router]
                    asn = update.asn
                    reason = reason_func(router, update)
                    f.write('node.AS\t{}\t{}\t{}\n'.format(name, asn, reason))

    def node_info(self, filename):
        with fopen2(filename, 'wt') as f:
            for name, router in self.bdrmapit.graph.routers.items():
                if name[0] == 'N':
                    update = self.rupdates[router]
                    if not router.hints:
                        hints = []
                    else:
                        hints = [int(h) for h in router.hints]
                    if isregexasn(update.utype):
                        restype = 'asn'
                    elif isregexorg(update.utype):
                        restype = 'org'
                    else:
                        restype = None
                    if restype is not None:
                        reasons = regexreasons(update.utype)
                    else:
                        reasons = []
                    origins = dict(Counter(interface.asn if interface.asn >= 0 else -100 for interface in router.interfaces))
                    succs = dict(Counter(interface.asn if interface.asn >= 0 else -100 for interface in router.succ))
                    dests = list(router.dests)
                    d = {'node': name, 'asn': int(update.asn), 'hints': hints, 'match': restype, 'reasons': reasons,
                         'origins': origins, 'subsequent': succs, 'dests': dests}
                    f.write(json.dumps(d) + '\n')

class Analyze:
    def __init__(self, bdrmapit: Bdrmapit):
        self.bdrmapit = bdrmapit

    def todf(self, interfaces: Collection[Interface], rupdates: Updates = None):
        if rupdates is None:
            rupdates = self.bdrmapit.rupdates
        rows = []
        for interface in interfaces:
            router = interface.router
            update = rupdates[router]
            asn = update.asn
            org = self.bdrmapit.as2org[asn]
            rtype = update.utype
            row = {'addr': interface.addr, 'asn': asn, 'org': org, 'rtype': rtype}
            rows.append(row)
        return pd.DataFrame(rows)

    def todf_addrs(self, addrs, *args, **kwargs):
        interfaces = [self.bdrmapit.graph.interfaces[addr] for addr in addrs if addr in self.bdrmapit.graph.interfaces]
        return self.todf(interfaces, *args, **kwargs)

class Test:
    def __init__(self, bdrmapit, names, file, rupdates: Updates = None, iupdates: Updates = None):
        self.bdrmapit = bdrmapit
        self.rupdates = rupdates if rupdates is not None else bdrmapit.rupdates
        self.iupdates = iupdates if iupdates is not None else bdrmapit.iupdates
        self.names = names
        self.file = file

    def create_tags(self):
        interre = re.compile(r'([a-z]+)\..*\.ntwk\.msn\.net')
        tags = {}
        with open(self.file) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                tag, *asns = line.split()
                if asns:
                    tags[tag] = tuple({int(asn) for asn in asns})
        valmap = {}
        for addr, name in self.names.items():
            m = interre.match(name)
            if m:
                tag = m.group(1)
                if tag in tags:
                    valmap[addr] = tags[tag]
        self.valmap = valmap

    def todf(self, addrs=None):
        if addrs is not None:
            interfaces: Set[Interface] = {self.bdrmapit.graph.interfaces[a] for a in addrs if a in self.bdrmapit.graph.interfaces}
        else:
            interfaces: Set[Interface] = self.bdrmapit.graph.interfaces.values()
        values = []
        for interface in interfaces:
            addr = interface.addr
            router: Router = interface.router
            rupdate: UpdateObj = self.rupdates[router]
            iupdate: UpdateObj = self.iupdates[interface]
            if rupdate is None:
                rasn = -1
                rorg = -1
                rtype = -1
            else:
                rasn = rupdate.asn
                rorg = rupdate.org
                rtype = rupdate.utype
            if iupdate is None or interface.org != rorg:
                iasn = interface.asn
                iorg = interface.org
                itype = -1 if iupdate is None else 0
            else:
                iasn = iupdate.asn
                iorg = iupdate.org
                itype = iupdate.utype
            row = {'addr': addr, 'router': router.name, 'asn': rasn, 'org': rorg, 'conn_asn': iasn, 'conn_org': iorg,
                   'rtype': rtype, 'itype': itype}
            values.append(row)
        return pd.DataFrame(values)

    def ixps(self, start_asn=None):
        values = []
        seen = set()
        for router in self.bdrmapit.routers_succ:
            conn_asn = self.rupdates[router].asn
            if conn_asn != start_asn:
                continue
            conn_org = self.bdrmapit.as2org[conn_asn]
            for isucc in router.succ:
                if isucc.asn <= -100:
                    if isucc.addr in seen:
                        continue
                    seen.add(isucc.addr)
                    pid = (isucc.asn * -1) - 100
                    rsucc = isucc.router
                    asn = self.rupdates[rsucc].asn
                    org = self.bdrmapit.as2org[asn]
                    value = {'addr': isucc.addr, 'router': router.name, 'asn': asn, 'org': org, 'conn_asn': conn_asn, 'conn_org': conn_org, 'pid': pid}
                    values.append(value)
        return pd.DataFrame(values)



from itertools import filterfalse, groupby
from operator import itemgetter
from libc.math cimport NAN, isnan, INFINITY


cpdef list max_num(iterable, key=None):
    cdef:
        list max_items = []
        float max_value = float('-inf')
        float value
    for item, value in zip(iterable, (map(key, iterable) if key else iterable)):
        if value > max_value:
            max_items = [item]
            max_value = value
        elif value == max_value:
            max_items.append(item)
    return max_items


cpdef tuple max2(iterable, key=None):
    cdef:
        float first_value = -INFINITY
        float second_value = -INFINITY
        float n
    first = None
    second = None
    for v in iterable:
        n = key(v) if key is not None else v
        if n > first_value:
            second = first
            second_value = first_value
            first = v
            first_value = n
        elif n > second_value:
            second = v
            second_value = n
    return first, first_value, second, second_value

cpdef tuple max2_values(iterable):
    cdef:
        float first_value = -INFINITY
        float second_value = -INFINITY
        float n
    first = None
    second = None
    for v, n in iterable.items():
        if n > first_value:
            second = first
            second_value = first_value
            first = v
            first_value = n
        elif n > second_value:
            second = v
            second_value = n
    return first, first_value, second, second_value

def unique_everseen(iterable, key=None):
    cdef set seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def unique_single_element(iterable, key=None):
    seen = set()
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            if len(seen) > 0:
                return False
            seen.add(element)
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                if len(seen) > 0:
                    return False
                seen.add(k)
    return True


def read_filenames(filename):
    with open(filename) as f:
        for line in f:
            if not line.startswith('#'):
                line = line.strip()
                if line:
                    yield line


cpdef peek(s):
    for i in s:
        return i


def unique_justseen(iterable, key=None):
    """List unique elements, preserving order. Remember only the element just seen."""
    # unique_justseen('AAAABBBCCDAABBB') --> A B C D A B
    # unique_justseen('ABBCcAD', str.lower) --> A B C A D
    return map(next, map(itemgetter(1), groupby(iterable, key)))


from file2 import fopen

from traceutils2.radix.radix cimport Radix
from traceutils2.radix.radix_node cimport RadixNodeASNs
from traceutils2.radix.radix_tree cimport RadixTreeASNs
from libc.stdlib cimport atol

cdef list PRIVATE4 = ['0.0.0.0/8', '10.0.0.8/8', '100.64.0.0/10', '127.0.0.0/8', '169.254.0.0/16', '172.16.0.0/12',
            '192.0.0.0/24', '192.0.2.0/24', '192.31.196.0/24', '192.52.193.0/24', '192.88.99.0/24', '192.168.0.0/16',
            '192.175.48.0/24', '198.18.0.0/15', '198.51.100.0/24', '203.0.113.0/24', '240.0.0.0/4',
            '255.255.255.255/32']
cdef list PRIVATE6 = ['::1/128', '::/128', '::ffff:0:0/96', '64:ff9b::/96', '100::/64', '2001::/23', '2001::/32', '2001:1::1/128',
            '2001:2::/48', '2001:3::/32', '2001:4:112::/48', '2001:5::/32', '2001:10::/28', '2001:20::/28',
            '2001:db8::/32', '2002::/16', '2620:4f:8000::/48', 'fc00::/7', 'fe80::/10']
cdef str MULTICAST4 = '224.0.0.0/3'
cdef str MULTICAST6 = 'FF00::/8'

cdef class IP2ASes(Radix):

    def __init__(self):
        tree4 = RadixTreeASNs()
        tree6 = RadixTreeASNs()
        super().__init__(tree4, tree6)

    def __getitem__(self, str item):
        return self.asns(item)

    cpdef RadixNodeASNs add_asns(self, str network, short masklen=-1, list asns=None):
        cdef RadixNodeASNs node = self.add(network, masklen)
        node.asns = asns
        return node

    cpdef list asns(self, str addr):
        cdef RadixNodeASNs node = self.search_best(addr)
        if node:
            return node.asns
        return []

    cpdef list asns_packed(self, bytes packed):
        cdef RadixNodeASNs node = self.search_best_packed(packed, -1)
        if node:
            return node.asns
        return 0

    cpdef void add_private(self) except *:
        for prefix in PRIVATE4:
            self.add_asns(prefix, -1, [-1])
        for prefix in PRIVATE6:
            self.add_asns(prefix, -1, [-1])
        self.add_asns(MULTICAST4, -1, [-1])
        self.add_asns(MULTICAST6, -1, [-1])


cpdef IP2ASes create_private():
    cdef IP2ASes ip2as = IP2ASes()
    ip2as.add_private()
    return ip2as


cpdef IP2ASes create_table(str filename):
    cdef bytes prefix, asns_str, asn_str
    cdef long asn
    cdef str prefix_s
    cdef list asns
    cdef IP2ASes ip2as = IP2ASes()
    with fopen(filename, 'rb') as f:
        # f.readline()
        for line in f:
            asns = []
            prefix, asns_str = line.rstrip().split()
            for asn_str in asns_str.split(b','):
                asn = atol(asn_str)
                asns.append(asn)
            prefix_s = prefix.decode()
            ip2as.add_asns(prefix_s, -1, asns=asns)
    return ip2as



from traceutils2.radix.radix_node cimport RadixNode, RadixNodeASN, RadixNodeASNs, RadixNodeData
from traceutils2.radix.radix_prefix cimport RadixPrefix


cdef unsigned char addr_test(bytes addr, unsigned char bitlen);
cdef bint prefix_match(RadixPrefix left, RadixPrefix right, unsigned char bitlen) except -1;
# cdef bint prefix_match2(bytes left, bytes right, unsigned char bitlen);

cdef class RadixTree:
    # cdef unsigned char maxbits
    cdef RadixNode head
    cdef long active_nodes

    cdef RadixNode create_node(self, RadixPrefix prefix=*, unsigned char prefix_size=*, RadixNode parent=*);
    cdef RadixNode add(self, RadixPrefix prefix);
    cdef void remove(self, RadixNode node) except *;
    cdef RadixNode search_best(self, RadixPrefix prefix);
    cdef RadixNode search_exact(self, RadixPrefix prefix);
    cdef RadixNode search_worst(self, RadixPrefix prefix);
    cdef list search_covered(self, RadixPrefix prefix);

cdef class RadixTreeASN(RadixTree):
    cdef RadixNodeASN add_asn(self, RadixPrefix prefix, int asn);

cdef class RadixTreeASNs(RadixTree):
    cdef RadixNodeASNs add_asns(self, RadixPrefix prefix, list asns);

cdef class RadixTreeData(RadixTree):
    cdef RadixNodeData add_data(self, RadixPrefix prefix, dict data=*)



from collections import defaultdict
from itertools import chain
from sys import stderr
from typing import Dict, Union

from file2 import fopen
from pb_amarder import Progress

from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Interface, Router
from bdrmapit.scripts.traceparser import ParseResults
from bdrmapit.vrf.vrfedge import VRFEdge

import pandas as pd

def construct_graph(ip2as, as2org, filename, remove_edges=None):
    prep = Container.load(ip2as, as2org, filename)
    if remove_edges is not None:
        for x, y in remove_edges:
            if (x, y) in prep.parseres.nextadjs:
                del prep.parseres.nextadjs[x, y]
            if (x, y) in prep.parseres.multiadjs:
                del prep.parseres.multiadjs[x, y]
    return prep.construct(no_echos=True)

class Container:
    def __init__(self, ip2as, as2org, parseres: ParseResults):
        self.ip2as = ip2as
        self.as2org = as2org
        self.parseres = parseres
        self.interfaces: Dict[str, Interface] = {}
        self.routers: Dict[str, Router] = {}
        self.addrs = None
        self.nexthops = None
        self.multi = None
        self.dps = None
        self.firstaddrs = None
        self.echos = None

    @classmethod
    def load(cls, ip2as, as2org, *files):
        allresults = None
        for file in files:
            results = ParseResults.load(file)
            if allresults is None:
                allresults = results
            else:
                allresults.update(results)
        return cls(ip2as, as2org, allresults)

    def alladdrs(self):
        return set(self.addrs) | set(self.parseres.echos)

    def filter_addrs(self, loop=True, no_echos=False):
        addrs = set()
        firstaddrs = {addr for _, addr in self.parseres.first}
        adjs = self.parseres.nextadjs + self.parseres.multiadjs
        pb = Progress(len(adjs), 'Filtering addrs', increment=1000000, callback=lambda: '{:,d}'.format(len(addrs)))
        for (x, y), n in pb.iterator(adjs.items()):
            if not loop or n > self.parseres.loopadjs.get((x, y), 0):
                addrs.add(x)
                addrs.add(y)
                firstaddrs.discard(y)
        self.addrs = addrs | firstaddrs
        self.addrs |= {addr for addr, _ in self.parseres.dps}
        if not no_echos:
            self.addrs |= self.parseres.echos
        # self.firstaddrs = {(file, addr) for file, addr in self.parseres.first if addr in firstaddrs}
        Progress.message('Total addrs: {:,d}'.format(len(self.addrs)), file=stderr)

    def set_echos(self):
        pass

    def create_edges(self, loop=True):
        nexthops = defaultdict(set)
        kept = 0
        pb = Progress(len(self.parseres.nextadjs), increment=200000, callback=lambda: '{:,d}'.format(kept))
        for (x, y), n in pb.iterator(self.parseres.nextadjs.items()):
            if x != y:
                if not loop or n + self.parseres.multiadjs.get((x, y), 0) > self.parseres.loopadjs.get((x, y), 0):
                    xasn = self.ip2as[x]
                    yasn = self.ip2as[y]
                    if xasn == yasn or n > self.parseres.multiadjs.get((x, y), 0):
                        nexthops[x].add(y)
                        kept += 1
        nkept = kept
        mkept = 0
        multi = defaultdict(set)
        pb = Progress(len(self.parseres.multiadjs), increment=200000, callback=lambda: 'N {:,d} M {:,d}'.format(nkept, mkept))
        for (x, y), n in pb.iterator(self.parseres.multiadjs.items()):
            if x != y:
                if not loop or n + self.parseres.nextadjs.get((x, y), 0) > self.parseres.loopadjs.get((x, y), 0):
                    xasn = self.ip2as[x]
                    yasn = self.ip2as[y]
                    if xasn > 0 and yasn > 0 and xasn == yasn:
                        nexthops[x].add(y)
                        nkept += 1
                    elif x not in nexthops:
                        multi[x].add(y)
                        mkept += 1
        nexthops.default_factory = None
        multi.default_factory = None
        self.nexthops = nexthops
        self.multi = multi

    def create_dps(self):
        dps = defaultdict(set)
        pb = Progress(len(self.parseres.dps), 'Creating dest pairs', increment=1000000)
        for addr, asn in pb.iterator(self.parseres.dps):
            if asn > 0:
                dps[addr].add(asn)
        dps.default_factory = None
        self.dps = dps

    def create_node(self, addr, router: Router):
        """
        Create new interface node and assign it to router.
        :param addr: address of new interface node
        :param router: router node representing the interface's router
        """
        asn = self.ip2as.asn(addr)
        # Make sure address is not from private address space
        if asn >= 0 or asn <= -100:
            # Create interface
            interface = Interface(addr, asn, self.as2org[asn])
            self.interfaces[addr] = interface
            interface.router = router
            # Add interface to router
            router.interfaces.append(interface)
            self.routers[router.name] = router
            # interface.echo = echo
            # interface.cycle = cycle

    def create_nodes(self, nodes_file, no_echos: bool = False, increment=100000):
        """
        Create router nodes based on alias resolution.
        :param no_echos: ignore echo-only addresses
        :param nodes_file: filename containing alias resolution groupings in CAIDA format
        :param increment: increment for status
        """
        if not no_echos:
            taddrs = self.alladdrs()
        else:
            taddrs = self.addrs
        pb = Progress(message='Creating nodes', increment=increment, callback=lambda: 'Routers {:,d} Interfaces {:,d}'.format(len(self.routers), len(self.interfaces)))
        with fopen(nodes_file, 'rt') as f:
            for line in pb.iterator(f):
                if line[0] != '#':
                    _, nid, *naddrs = line.split()
                    if not any(addr in taddrs for addr in naddrs):
                        continue
                    nid = nid[:-1]
                    router = Router(nid)
                    self.routers[router.name] = router
                    for addr in naddrs:
                        self.create_node(addr, router)

    def create_remaining(self, aliases: bool, no_echos: bool = False, increment=100000):
        """
        Create router nodes for any interfaces not seen in the alias resolution dataset, or when there is not alias resolution dataset.
        :param aliases: flag to indicate if alias resoultion was used
        :param increment: increment for status
        """
        if not no_echos:
            taddrs = chain(self.addrs, self.parseres.echos)
            num_addrs = len(self.addrs) + len(self.parseres.echos)
        else:
            taddrs = self.addrs
            num_addrs = len(self.addrs)
        pb = Progress(num_addrs, 'Creating remaining routers and interfaces', increment=increment)
        for addr in pb.iterator(taddrs):
            if not aliases or addr not in self.interfaces:
                router = Router(addr)
                self.create_node(addr, router)

    @staticmethod
    def add_succ(router: Router, interface: Interface, succ: Union[VRFEdge, Interface]):
        if succ in router.succ:
            origins = router.origins[succ]
            origins.add(interface.asn)
        else:
            router.succ.add(succ)
            router.origins[succ] = {interface.asn}

    @staticmethod
    def add_pred(interface: Interface, prouter: Router):
        predcount = interface.pred.get(prouter, 0)
        interface.pred[prouter] = predcount + 1

    def add_nexthop(self, increment=100000):
        """
        Add nexthop edges.
        :param nexthop: nexthop edges
        :param increment: increment for status
        """
        pb = Progress(len(self.nexthops), 'Adding nexthop edges', increment=increment)
        for addr, edges in pb.iterator(self.nexthops.items()):
            interface = self.interfaces[addr]
            router = interface.router
            router.nexthop = True
            for edge in edges:
                succ = self.interfaces[edge]
                if succ.router != router:
                    self.add_succ(router, interface, succ)
                    self.add_pred(succ, router)

    def add_multi(self, increment=100000):
        """
        Add multiple hop edges.
        :param multi: multiple hop edges
        :param increment: increment for status
        """
        pb = Progress(len(self.multi), 'Adding multihop edges', increment=increment)
        for addr in pb.iterator(self.multi):
            interface = self.interfaces[addr]
            router = interface.router
            if not router.nexthop:
                edges = self.multi[addr]
                for edge in edges:
                    succ = self.interfaces[edge]
                    if succ.router != router:
                        self.add_succ(router, interface, succ)

    def add_dests(self, increment=100000):
        """
        Add destination ASes for each interface and router.
        :param dps: interface to destination mappings
        :param increment: increment for status
        """
        pb = Progress(len(self.dps), 'Adding destination ASes', increment=increment)
        for addr, dests in pb.iterator(self.dps.items()):
            interface = self.interfaces.get(addr)
            if interface is not None:
                interface.dests.update(dests)

    def create_graph(self):
        """
        Create the graph based on the interfaces, routers, and edges.
        :return: the graph
        """
        return Graph(interfaces=self.interfaces, routers=self.routers)

    def reset_hints(self):
        for interface in self.interfaces.values():
            interface.hint = 0
            interface.router.hints = None

    def add_hints(self, hints: Dict[str, int]):
        for addr, hint in hints.items():
            if addr in self.interfaces:
                interface = self.interfaces[addr]
                interface.hint = hint
                if not interface.router.hints:
                    interface.router.hints = {hint}
                else:
                    interface.router.hints.add(hint)

    def add_hints_file(self, filename):
        print('Adding hints from {}'.format(filename))
        df = pd.read_csv(filename, sep=r'\s+', index_col=0, names=['addr', 'tasn'])
        hints = dict(df.tasn)
        self.add_hints(hints)

    def reset(self):
        self.interfaces = {}
        self.routers = {}

    def construct(self, nodes_file=None, loop=True, hints_file=None, no_echos=False):
        """
        Construct the graph from scratch.
        :param addrs: addresses seen in the dataset
        :param nexthop: nexthop edges
        :param multi: multiple hop edges
        :param dps: interface to destination ASes
        :param nodes_file: alias resolution dataset
        :return: the graph
        """
        self.filter_addrs(loop=loop, no_echos=no_echos)
        self.create_edges(loop=loop)
        self.create_dps()
        if nodes_file is not None:
            self.create_nodes(nodes_file=nodes_file, no_echos=no_echos)
        self.create_remaining(nodes_file is not None, no_echos=no_echos)
        self.add_nexthop()
        self.add_multi()
        self.add_dests()
        if hints_file is not None:
            self.add_hints_file(hints_file)
        return self.create_graph()



from traceutils2.radix.ip2as cimport IP2AS
from traceutils2.utils.net cimport inet_pton_auto_str

cpdef ICMPType gettype(int family, int icmp_type, int icmp_code) except *:
    if family == AF_INET:
        if icmp_type == 0:
            return ICMPType.echo_reply
        elif icmp_type == 3:
            if icmp_code == 13:
                return ICMPType.spoofing
            elif icmp_code == 3:
                return ICMPType.portping
            elif icmp_code == 0:
                return ICMPType.net_unreach
            elif icmp_code == 1:
                return ICMPType.host_unreach
            else:
                return ICMPType.dest_unreach
        elif icmp_type == 8:
            return ICMPType.echo_request
        elif icmp_type == 11:
            return ICMPType.time_exceeded
    else:
        if icmp_type == 129:
            return ICMPType.echo_reply
        elif icmp_type == 1:
            if icmp_code == 1:
                return ICMPType.spoofing
            elif icmp_code == 4:
                return ICMPType.portping
            else:
                return ICMPType.dest_unreach
        elif icmp_type == 128:
            return ICMPType.echo_request
        elif icmp_type == 3:
            return ICMPType.time_exceeded

cdef class Hop:

    def __init__(self, str addr, unsigned char probe_ttl, double rtt, unsigned char reply_ttl, int reply_tos, int reply_size, unsigned char icmp_type, unsigned char icmp_code, unsigned char icmp_q_ttl, int icmp_q_tos, int family):
        self.addr = addr
        self.probe_ttl = probe_ttl
        self.rtt = rtt
        self.reply_ttl = reply_ttl
        self.reply_tos = reply_tos
        self.reply_size = reply_size
        self.icmp_type = icmp_type
        self.icmp_code = icmp_code
        self.icmp_q_ttl = icmp_q_ttl
        self.icmp_q_tos = icmp_q_tos
        self.family = family

    def __repr__(self):
        return '{ttl:02d}: {addr}'.format(addr=self.addr, ttl=self.probe_ttl)

    cpdef bytes set_packed(self):
        self.packed = inet_pton_auto_str(self.addr)
        return self.packed


cdef class Trace:

    def __init__(self, str src, str dst, list hops):
        self.src = src
        self.dst = dst
        self.allhops = hops
        self.hops = hops
        self.loop = None
        self.family = 0
        self.jdata = None

    def __iter__(self):
        for h in self.hops:
            yield h

    def __repr__(self):
        return '\n'.join(repr(hop) for hop in self.hops)

    cpdef list addrs(self):
        cdef Hop h
        return [h.addr for h in self.hops]

    cpdef void prune_dups(self) except *:
        cdef str prev = None, haddr
        cdef list hops = []
        # cdef int i
        cdef Hop hop
        # for i in range(len(self.hops)):
        #     hop = self.hops[i]
        for hop in self.hops:
            haddr = hop.addr
            if haddr != prev:
                hops.append(hop)
                prev = haddr
            else:
                hops[-1] = hop
        self.hops = hops

    cpdef unsigned char mark_loop(self) except? -1:
        cdef str addr, prev
        cdef set seen = set()
        cdef unsigned char end = len(self.hops), i
        prev = None
        for i in range(len(self.hops) - 1, -1, -1):
            addr = self.hops[i].addr
            if addr in seen and addr != prev:
                end = i
            else:
                seen.add(addr)
            prev = addr
        if end < len(self.hops):
            return end
        return 255

    # cpdef void prune_loops(self, bint keepfirst=False) except *:
    #     cdef set seen = set()
    #     cdef int end = len(self.hops), i
    #     cdef str addr, prev
    #     # cdef Hop poss_end1, poss_end2
    #     prev = None
    #     for i in range(len(self.hops) - 1, -1, -1):
    #         addr = self.hops[i].addr
    #         if addr in seen and addr != prev:
    #             end = i
    #         else:
    #             seen.add(addr)
    #         prev = addr
    #     if end < len(self.hops):
    #         # poss_end1 = self.hops[end]
    #         # poss_end2 = self.hops[end+1]
    #         self.loop = self.hops[end:]
    #         self.loopstart = end
    #         if keepfirst or len(self.loop) <= 3:
    #             self.hops = self.hops[:end+1]
    #         else:
    #             self.hops = self.hops[:end]
    #         # if poss_end1.reply_ttl >= poss_end2.reply_ttl:
    #         #     self.hops = self.hops[:end+1]
    #         # else:
    #         #     self.hops = self.hops[:end]
    cpdef void prune_loops(self, bint keepfirst=False) except *:
        cdef unsigned char end = self.mark_loop()
        self.loop = self.hops[end:]
        if keepfirst or len(self.loop) <= 3:
            self.hops = self.hops[:end+1]
        else:
            self.hops = self.hops[:end]
        # if poss_end1.reply_ttl >= poss_end2.reply_ttl:
        #     self.hops = self.hops[:end+1]
        # else:
        #     self.hops = self.hops[:end]

    cpdef void prune_private(self, IP2AS ip2as) except *:
        cdef Hop h
        self.hops = [h for h in self.hops if ip2as[h.addr] != -1]

    cpdef void prune_src(self, str src2=None) except *:
        cdef Hop h
        self.hops = [h for h in self.hops if h.addr != self.src and h.addr != src2]

    cpdef void set_packed(self) except *:
        for hop in self.hops:
            hop.set_packed()


cdef class Reader:

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def raw(self):
        raise NotImplementedError()

    cpdef void open(self) except *:
        raise NotImplementedError()

    cpdef void close(self) except *:
        raise NotImplementedError()



from socket import AF_INET

from traceutils2.radix.radix_tree cimport RadixTree, RadixTreeASN
from traceutils2.radix.radix_node cimport RadixNode
from traceutils2.radix.radix_prefix cimport from_packed, from_network, from_address, RadixPrefix


def _iter(RadixNode node):
    cdef list stack
    stack = []
    while node is not None:
        if node.prefix:
            yield node
        if node.left:
            if node.right:
                # we'll come back to it
                stack.append(node.right)
            node = node.left
        elif node.right:
            node = node.right
        elif len(stack) != 0:
            node = stack.pop()
        else:
            break


cdef list _search_covering(RadixNode node):
    cdef list stack
    stack = []
    while node is not None:
        if node.prefix:
            stack.append(node)
        node = node.parent
    return stack


cdef class Radix:

    def __init__(self, RadixTree tree4=None, RadixTree tree6=None):
        self._tree4 = tree4 if tree4 is not None else RadixTree()
        self._tree6 = tree6 if tree6 is not None else RadixTree()
        self.gen_id = 0            # detection of modifiction during iteration

    def __iter__(self):
        cdef long init_id
        cdef RadixNode elt
        init_id = self.gen_id
        for elt in _iter(self._tree4.head):
            if init_id != self.gen_id:
                raise RuntimeWarning('detected modification during iteration')
            yield elt
        for elt in _iter(self._tree6.head):
            if init_id != self.gen_id:
                raise RuntimeWarning('detected modification during iteration')
            yield elt

    cdef RadixNode _add(self, RadixPrefix prefix):
        cdef RadixNode node
        if prefix.family == AF_INET:
            node = self._tree4.add(prefix)
        else:
            node = self._tree6.add(prefix)
        self.gen_id += 1
        return node

    cdef void _delete(self, RadixNode node) except *:
        if not node:
            raise KeyError('match not found')
        if node.prefix.family == AF_INET:
            self._tree4.remove(node)
        else:
            self._tree6.remove(node)
        self.gen_id += 1

    cdef RadixNode _search_exact(self, RadixPrefix prefix):
        if prefix.family == AF_INET:
            return self._tree4.search_exact(prefix)
        else:
            return self._tree6.search_exact(prefix)

    cdef RadixNode _search_best(self, RadixPrefix prefix):
        if prefix.family == AF_INET:
            return self._tree4.search_best(prefix)
        else:
            return self._tree6.search_best(prefix)

    cdef RadixNode _search_worst(self, RadixPrefix prefix):
        if prefix.family == AF_INET:
            return self._tree4.search_worst(prefix)
        else:
            return self._tree6.search_worst(prefix)

    cdef list _search_covered(self, RadixPrefix prefix):
        if prefix.family == AF_INET:
            return self._tree4.search_covered(prefix)
        return self._tree6.search_covered(prefix)

    cpdef RadixNode add(self, str network, short masklen=-1):
        return self._add(from_network(network, masklen))

    cpdef RadixNode add_packed(self, bytes packed, unsigned char masklen):
        return self._add(from_packed(packed, masklen))

    cpdef void delete(self, str network, short masklen=-1) except *:
        self._delete(self.search_exact_prefix(network, masklen))

    cpdef void delete_packed(self, bytes packed, unsigned char masklen) except *:
        self._delete(self.search_exact_packed(packed, masklen))

    cpdef RadixNode search_exact(self, str addr):
        return self._search_exact(from_address(addr))

    cpdef RadixNode search_exact_prefix(self, str network, short masklen=-1):
        return self._search_exact(from_network(network, masklen))

    cpdef RadixNode search_exact_packed(self, bytes packed, unsigned char masklen):
        return self._search_exact(from_packed(packed, masklen))

    cpdef RadixNode search_best(self, str addr):
        return self._search_best(from_address(addr))

    cpdef RadixNode search_best_prefix(self, str network, short masklen=-1):
        return self._search_best(from_network(network, masklen))

    cpdef RadixNode search_best_packed(self, bytes packed, unsigned char masklen):
        return self._search_best(from_packed(packed, masklen))

    cpdef RadixNode search_worst(self, str addr):
        return self._search_worst(from_address(addr))

    cpdef RadixNode search_worst_prefix(self, str network, short masklen=-1):
        return self._search_worst(from_network(network, masklen))

    cpdef RadixNode search_worst_packed(self, bytes packed, unsigned char masklen):
        return self._search_worst(from_packed(packed, masklen))

    cpdef list search_covered(self, str addr):
        return self._search_covered(from_address(addr))

    cpdef list search_covered_prefix(self, str network, short masklen=-1):
        return self._search_covered(from_network(network, masklen))

    cpdef list search_covered_packed(self, bytes packed, unsigned char masklen):
        return self._search_covered(from_packed(packed, masklen))

    cpdef list search_covering(self, str addr):
        return _search_covering(self.search_best(addr))

    cpdef list search_covering_prefix(self, str network, short masklen=-1):
        return _search_covering(self.search_best_prefix(network, masklen))

    cpdef list search_covering_packed(self, bytes packed, unsigned char masklen):
        return _search_covering(self.search_best_packed(packed, masklen))

    cpdef list nodes(self):
        cdef RadixNode elt
        return [elt for elt in self]

    cpdef list prefixes(self):
        cdef RadixNode elt
        return [str(elt.prefix) for elt in self]



from traceutils2.radix.ip2as cimport IP2AS
cdef extern from '<arpa/inet.h>':
    cdef int AF_INET, AF_INET6

cpdef enum TraceFType:
    none = 0
    warts = 1
    wartsjson = 2
    atlas = 3

cpdef enum ICMPType:
    echo_reply = 1
    dest_unreach = 2
    ptb = 3
    time_exceeded = 4
    echo_request = 5
    spoofing = 6
    portping = 7
    net_unreach = 8
    host_unreach = 9

cpdef ICMPType gettype(int family, int icmp_type, int icmp_code) except *;

cdef class Hop:
    cdef:
        public str addr
        public unsigned char probe_ttl
        public double rtt
        public unsigned char reply_ttl
        public int reply_tos
        public int reply_size
        public unsigned char icmp_type
        public unsigned char icmp_code
        public unsigned char icmp_q_ttl
        public int icmp_q_tos
        public bytes packed
        public bint ismpls
        public int family
        public ICMPType type

    cpdef bytes set_packed(self);


cdef class Trace:
    cdef:
        public str src
        public str dst
        public list hops, allhops
        public list loop
        public int family
        public str jdata

    cpdef list addrs(self);
    cpdef unsigned char mark_loop(self) except? -1;
    cpdef void prune_dups(self) except *;
    cpdef void prune_loops(self, bint keepfirst=*) except *;
    cpdef void prune_private(self, IP2AS ip2as) except *
    cpdef void prune_src(self, str src2=*) except *;
    cpdef void set_packed(self) except *;

cdef class Reader:
    cdef:
        public str filename
        public p

    cpdef void open(self) except *;
    cpdef void close(self) except *;

# cpdef unsigned long iptoint(str addr) except -1;


from abc import ABC, abstractmethod
from typing import Optional

from bdrmapit.algorithm.updates_dict import Updates
from bdrmapit.graph.construct import Graph
from bdrmapit.graph.node import Router, Interface

DEBUG = False

class Debug:

    def __init__(self, bdrmapit=None, rupdates=None, iupdates=None, verbose=True):
        self.old = DEBUG
        self.bdrmapit = bdrmapit
        self.rupdates = rupdates
        self.iupdates = iupdates
        self.old_rupdates = None
        self.old_iupdates = None
        self.verbose = verbose

    def __enter__(self):
        global DEBUG
        DEBUG = self.verbose
        if self.bdrmapit is not None:
            if self.rupdates is not None:
                self.old_rupdates, self.bdrmapit.rupdates = self.bdrmapit.rupdates, self.rupdates
            if self.iupdates is not None:
                self.old_iupdates, self.bdrmapit.iupdates = self.bdrmapit.iupdates, self.iupdates

    def __exit__(self, exc_type, exc_val, exc_tb):
        global DEBUG
        DEBUG = self.old
        if self.bdrmapit is not None:
            if self.rupdates is not None:
                self.bdrmapit.rupdates = self.old_rupdates
            if self.iupdates is not None:
                self.bdrmapit.iupdates = self.old_iupdates
        return False

class DebugMixin(ABC):

    def __init__(self):
        self.graph: Optional[Graph] = None

    @abstractmethod
    def annotate_lasthop(self, router: Router):
        raise NotImplementedError()

    @abstractmethod
    def annotate_router(self, router: Router, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def annotate_router_vrf(self, router: Router):
        raise NotImplementedError()

    @abstractmethod
    def annotate_interface(self, interface: Interface):
        raise NotImplementedError()

    @abstractmethod
    def annotate_router_hint(self, router: Router):
        raise NotImplementedError()

    def test_last(self, nid, rupdates=None, iupdates=None, usehints=False):
        if rupdates is None:
            rupdates = Updates()
        if iupdates is None:
            iupdates = Updates()
        with Debug(self, rupdates=rupdates, iupdates=iupdates):
            try:
                r: Router = self.graph.routers[nid]
            except KeyError:
                r: Router = self.graph.interfaces[nid].router
            if usehints:
                asn, utype = self.annotate_router_hint(r)
                if asn > 0:
                    return asn, utype
            result = self.annotate_lasthop(r)
        print(result)

    def test_router(self, nid, rupdates=None, iupdates=None, usehints=False, **kwargs):
        with Debug(self, rupdates=rupdates, iupdates=iupdates):
            if isinstance(nid, Router):
                r: Router = nid
            else:
                try:
                    r: Router = self.graph.routers[nid]
                except KeyError:
                    r: Router = self.graph.interfaces[nid].router
            if usehints:
                asn, utype = self.annotate_router_hint(r, **kwargs)
                if asn > 0:
                    return asn, utype
            if r.vrf:
                result = self.annotate_router_vrf(r)
            else:
                result = self.annotate_router(r, **kwargs)
        print(result)
        return result

    def test_interface(self, addr, rupdates=None, iupdates=None):
        with Debug(self, rupdates=rupdates, iupdates=iupdates):
            i = self.graph.interfaces[addr]
            result = self.annotate_interface(i)
        print(result)



from file2 import fopen

from traceutils2.utils.dicts cimport StrDict, EmptyDict

cdef class AS2Org:

    def __init__(self, str filename, str additional=None):
        cdef dict org_names, siblings
        cdef set v

        self.orgs = StrDict()
        self.asn_names = StrDict()
        self.asn_org_names = StrDict()
        self.siblings = EmptyDict()
        self.org_cc = {}
        org_names = {}
        read_org = False
        read_asn = False
        with fopen(filename) as f:
            for line in f:
                line = line.strip()
                # if line == '# format:org_id|changed|org_name|country|source':
                if line.startswith('# format:org_id|changed|org_name|country'):
                    read_org = True
                    read_asn = False
                # elif line == '# format:aut|changed|aut_name|org_id|opaque_id|source':
                elif line.startswith('# format:aut|changed|aut_name|org_id'):
                    read_asn = True
                    read_org = False
                elif read_org:
                    org_id, changed, org_name, country, *extra = line.split('|')
                    org_names[org_id] = org_name
                    self.org_cc[org_id] = country
                elif read_asn:
                    aut, changed, aut_name, org_id, *extra = line.split('|')
                    asn = int(aut)
                    self.orgs[asn] = org_id
                    self.asn_names[asn] = aut_name
        for asn, org_id in self.orgs.items():
            self.asn_org_names[asn] = org_names[org_id]
        self.set_siblings()
        if additional:
            with fopen(additional) as f:
                for line in f:
                    if line.strip():
                        if not line.startswith('#'):
                            asn, *splits = list(map(int, line.split()))
                            newsibs = set(splits)
                            newsibs -= self.siblings[asn]
                            newsibs = {s for sib in newsibs for s in self.siblings[sib]}
                            for other in newsibs:
                                self.orgs[other] = self.orgs[asn]
                                self.asn_org_names[other] = self.asn_org_names[asn]
            self.set_siblings()

    def set_siblings(self):
        siblings = {}
        for k in set(self.orgs.values()):
            siblings[k] = set()
        for asn, org_id in self.orgs.items():
            siblings[org_id].add(asn)
        for v in siblings.values():
            for asn in v:
                self.siblings[asn] = v

    def __getitem__(self, item):
        return self.orgs[item]

    cpdef str name(self, int asn):
        return self.asn_org_names[asn]

    cpdef str asn_name(self, int asn):
        return self.asn_names[asn]

    cpdef str asn_cc(self, int asn):
        return self.org_cc.get(self[asn], 'NA')



import pickle
from collections import defaultdict
from typing import Dict, Optional

from traceutils2.as2org.as2org import AS2Org
from traceutils2.radix.ip2as import IP2AS
from traceutils2.utils.net import otherside

from bdrmapit.container.container import Container
from scripts.traceparser import ParseResults
from bdrmapit.vrf.vrfedge import VRFEdge, VType


class VRFPrep(Container):
    def __init__(self, ip2as: IP2AS, as2org: AS2Org, parseres: ParseResults, vrfinfo=None):
        super().__init__(ip2as, as2org, parseres)
        if vrfinfo:
            self.middle = vrfinfo['middle']
            self.last = vrfinfo['last']
        else:
            self.middle = None
            self.last = None
        self.ip2as = ip2as
        self.create_edges()
        self.original_nexthop = dict(self.nexthops)
        self.original_multi = dict(self.multi)
        self.bnext: Optional[Dict[str, Dict[str, VType]]] = None
        self.anext = None
        self.bmulti: Optional[Dict[str, Dict[str, VType]]] = None
        self.amulti = None
        self.prune = None

    def load_vrfinfo(self, filename):
        with open(filename, 'rb') as f:
            vrfinfo = pickle.load(f)
        self.middle = vrfinfo['middle']
        self.last = vrfinfo['last']
        # self.middle = {x: asns for x, asns in self.middle.items() if asns != {self.ip2as[x]}}

    def merge_edgetypes(self, toforward, forwarding):
        bedges = defaultdict(dict)
        keys = toforward.keys() | forwarding.keys()
        for addr in keys:
            set1 = toforward[addr] if addr in toforward else set()
            set2 = forwarding[addr] if addr in forwarding else set()
            for succ in set1 - set2:
                bedges[addr][succ] = VType.toforward
            for succ in set2 - set1:
                bedges[addr][succ] = VType.forwarding
            for succ in set1 & set2:
                bedges[addr][succ] = VType.forwarding
        bedges.default_factory = None
        return bedges

    def mark_vrfs(self, triplets):
        toforward_next = defaultdict(set)
        forwarding_next = defaultdict(set)
        self.bnext = defaultdict(dict)
        self.anext = defaultdict(set)
        toforward_multi = defaultdict(set)
        forwarding_multi = defaultdict(set)
        self.bmulti = defaultdict(dict)
        self.amulti = defaultdict(set)
        self.prune = defaultdict(set)

        def mark(a, b, c=None):
            if a == '17.1.154.14' or b == '17.1.154.14':
                print(a, b, c)
            if a in self.original_nexthop and b in self.original_nexthop[a]:
                toforward = toforward_next
                forwarding = forwarding_next
                aedges = self.anext
            else:
                toforward = toforward_multi
                forwarding = forwarding_multi
                aedges = self.amulti
            if a:
                self.prune[a].add(b)
                toforward[a].add(b)
                forwarding[b].add(a)
            if c:
                self.prune[b].add(c)
                aedges[b].add(c)
                aedges[c].add(b)

        pb = Progress(message='Marking VRF edges', increment=500000, callback=lambda: '{:,d}'.format(len(self.prune)))
        print(self.middle.get('2001:504:0:1::6939:1'))
        with fopen(triplets) as f:
            for line in pb.iterator(f):
                w, x, y = line.split()
                if x in self.middle:
                    # if x == '2001:504:0:1::6939:1':
                    #     print(w, x, y)
                    if not w:
                        if None in self.middle[x]:
                            mark(x, y)
                    else:
                        if self.ip2as[w] in self.middle[x]:
                            mark(w, x, y)
                if y in self.last:
                    if not w:
                        if None not in self.last[y]:
                            mark(x, y)
                    else:
                        pasns = self.last[y]
                        if self.ip2as[x] not in pasns:
                            mark(x, y)
        self.bnext = self.merge_edgetypes(toforward_next, forwarding_next)
        self.bmulti = self.merge_edgetypes(toforward_multi, forwarding_multi)

    def remove_vrfs(self, edges, debug=None):
        nexthop = {}
        removed = set()
        pb = Progress(len(edges), 'Removing forwarding address edges', increment=500000, callback=lambda: 'K {:,d} R {:,d}'.format(len(nexthop), len(removed)))
        for x, succ in pb.iterator(edges.items()):
            if x == debug:
                print(x, succ, self.prune.get(x))
            if x in self.prune:
                newsucc = succ - self.prune[x]
                # newsucc = [y for y in succ if y not in self.prune[x]]
                if newsucc:
                    nexthop[x] = newsucc
                else:
                    removed.add(x)
            else:
                nexthop[x] = succ
        return nexthop

    def remove_nexthop(self, **kwargs):
        self.nexthops = self.remove_vrfs(self.original_nexthop, **kwargs)

    def remove_multi(self):
        self.multi = self.remove_vrfs(self.original_multi)

    def add_vrfedges(self, bedges: Dict[str, Dict[str, VType]], nexthop, skip_exists=True, increment=100000):
        etype = 'nexthop' if nexthop else 'multi'
        pb = Progress(len(bedges), 'Adding {} forwarding edges'.format(etype), increment=increment)
        for addr, succs in pb.iterator(bedges.items()):
            if addr not in self.interfaces:
                continue
            interface = self.interfaces[addr]
            router = interface.router
            if not nexthop and router.nexthop:
                continue
            if skip_exists and router.succ:
                continue
            router.nexthop = nexthop
            router.vrf = True
            for succ, vtype in succs.items():
                if succ not in self.interfaces:
                    continue
                srouter = self.interfaces[succ].router
                edge = VRFEdge(srouter, vtype)
                self.add_succ(router, interface, edge)

    def add_nexthop_forwarding(self, skip_exists=True, increment=100000):
        """
        Add nexthop forwarding edges.
        :param skip_exists: don't add vrf edges if normal edges exist
        :param increment: increment for status
        """
        self.add_vrfedges(self.bnext, True, skip_exists=skip_exists, increment=increment)

    def add_pred_forwarding(self, skip_exists=True, increment=100000):
        """
        Add backward forwarding edges.
        :param skip_exists: don't add vrf edges if normal edges exist
        :param increment: increment for status
        """
        pb = Progress(len(self.anext), 'Adding forwarding backward edges', increment=increment)
        for addr, edges in pb.iterator(self.anext.items()):
            if addr not in self.interfaces:
                continue
            interface = self.interfaces[addr]
            if skip_exists and interface.pred:
                continue
            for edge in edges:
                if edge not in self.interfaces:
                    continue
                try:
                    if otherside(edge, 2) == addr or otherside(edge, 4) == addr:
                        prouter = self.interfaces[edge].router
                        self.add_pred(interface, prouter)
                except:
                    pass

    def add_multi_forwarding(self, skip_exists=True, increment=100000):
        """
        Add multiple hop forwarding edges.
        :param skip_exists: don't add vrf edges if normal edges exist
        :param increment: increment for status
        """
        self.add_vrfedges(self.bmulti, False, skip_exists=skip_exists, increment=increment)

    def reset(self, keep_nodes=True):
        pb = Progress(len(self.routers), 'Resetting router edges', increment=1000000)
        for router in pb.iterator(self.routers.values()):
            router.succ.clear()
            router.origins.clear()
        pb = Progress(len(self.interfaces), 'Resetting interface edges', increment=1000000)
        for interface in pb.iterator(self.interfaces.values()):
            interface.pred.clear()

    def construct(self, nodes_file=None, skip_exists=True, skip_nodes=False, skip_dests=False, skip_graph=False):
        """
        Construct the graph from scratch.
        :param skip_graph: 
        :param skip_dests: 
        :param skip_nodes: 
        :param skip_exists: 
        :param nodes_file: alias resolution dataset
        :return: the graph
        """

        # self.filter_addrs()
        # self.create_edges()
        # self.create_dps()
        # if nodes_file is not None:
        #     self.create_nodes(nodes_file=nodes_file)
        # self.create_remaining(nodes_file is not None)
        # self.add_nexthop()
        # self.add_multi()
        # self.add_dests()
        # return self.create_graph()
        self.filter_addrs()
        self.create_dps()
        if not skip_nodes:
            if nodes_file is not None:
                self.create_nodes(nodes_file=nodes_file)
            self.create_remaining(nodes_file is not None)
        self.add_nexthop()
        self.add_nexthop_forwarding(skip_exists=skip_exists)
        self.add_pred_forwarding(skip_exists=skip_exists)
        self.add_multi()
        self.add_multi_forwarding(skip_exists=skip_exists)
        if not skip_dests:
            self.add_dests()
        if not skip_graph:
            return self.create_graph()
